# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015 Digi International Inc. All Rights Reserved.
import unittest

import six
from suitcase.fields import Magic, SBInt64, DispatchField, UBInt8, DispatchTarget, \
    LengthField
from suitcase.structure import Structure
from suitcase.protocol import StreamProtocolHandler
from suitcase.test.examples.test_network_stack import UDPFrame


class MagicSchema(Structure):
    magic = Magic(b'\xAA\xAA')
    value = SBInt64()

class LongMagicSchema(Structure):
    magic = Magic(b'\xAA\xAA\xBB\xBB')
    value = SBInt64()

class ErrorCaseSchema(Structure):
    type = DispatchField(UBInt8())
    length = LengthField(UBInt8())
    body = DispatchTarget(length, type, {
        0x00: MagicSchema
    })


class TestStreamProtocol(unittest.TestCase):
    def test_protocol_basic(self):
        packets_received = []

        def callback(packet):
            packets_received.append(packet)

        f = UDPFrame()
        f.data = b"Hello, world"
        f.checksum = 0x01
        f.destination_port = 9000
        f.source_port = 8000

        phandler = StreamProtocolHandler(UDPFrame, callback)
        packet_bytes = f.pack()
        rem_bytes = packet_bytes
        while len(rem_bytes) > 0:
            chunk = rem_bytes[:5]
            rem_bytes = rem_bytes[5:]
            phandler.feed(chunk)

        assert (len(packets_received) == 1)
        rx = packets_received[0]
        assert rx.data == b"Hello, world"
        assert rx.checksum == 0x01
        assert rx.destination_port == 9000
        assert rx.source_port == 8000

    def test_protocol_violation(self):
        # verify that when there is a protocol violation, that state gets
        # cleaned up properly.  For this, we have a dispatch field that
        # has an unmapped type byte

        rx = []

        def bad_cb(packet):
            self.fail("Callback should not have been executed")

        def good_cb(packet):
            rx.append(packet)
            self.assertEqual(packet.body.value, -31415926)

        good_msg = ErrorCaseSchema()
        good_msg.body = MagicSchema()
        good_msg.body.value = -31415926
        good_packet = good_msg.pack()

        bad_packet = b'\x01\x02\xAA\xAA' + good_packet
        phandler = StreamProtocolHandler(ErrorCaseSchema, bad_cb)
        phandler.feed(bad_packet)

        phandler.packet_callback = good_cb
        phandler.feed(good_packet)

        self.assertEqual(len(rx), 1)

    def test_protocol_magic_scan_full_buffer(self):
        rx = []

        def cb(packet):
            rx.append(packet)

        protocol_handler = StreamProtocolHandler(MagicSchema, cb)
        test_sequence = MagicSchema()
        test_sequence.value = -29939
        pack = test_sequence.pack()

        # garbage with our bytes in the middle
        test_bytes = b'\x1A\x3fadbsfkasdf;aslkfjasd;f' + pack + b'\x00\x00asdfn234r'
        protocol_handler.feed(test_bytes)
        self.assertEqual(len(rx), 1)
        self.assertEqual(rx[0].value, -29939)

    def test_protocol_magic_scan_two_part(self):
        rx = []

        def cb(packet):
            rx.append(packet)

        protocol_handler = StreamProtocolHandler(MagicSchema, cb)
        test_sequence = MagicSchema()
        test_sequence.value = -29939
        pack = test_sequence.pack()

        # garbage with our bytes in the middle
        test_bytes = b'\x1A\x3fadbsfkasdf;aslkfjasd;f' + pack[:1]
        test_bytes2 = pack[1:] + b'\x00\x00asdfn234r'
        protocol_handler.feed(test_bytes)
        protocol_handler.feed(test_bytes2)
        self.assertEqual(len(rx), 1)
        self.assertEqual(rx[0].value, -29939)

    def test_protocol_magic_scan_two_part_longer(self):
        rx = []

        def cb(packet):
            rx.append(packet)

        protocol_handler = StreamProtocolHandler(LongMagicSchema, cb)
        test_sequence = LongMagicSchema()
        test_sequence.value = -29939
        pack = test_sequence.pack()

        # garbage with our bytes in the middle
        test_bytes = b'\x1A\x3fadbsfkasdf;aslkfjasd;f' + pack[:2]
        test_bytes2 = pack[2:] + b'\x00\x00asdfn234r'
        protocol_handler.feed(test_bytes)
        protocol_handler.feed(test_bytes2)
        self.assertEqual(len(rx), 1)
        self.assertEqual(rx[0].value, -29939)

    def test_protocol_magic_scan_single_byte(self):
        rx = []

        def cb(packet):
            rx.append(packet)

        protocol_handler = StreamProtocolHandler(MagicSchema, cb)
        test_sequence = MagicSchema()
        test_sequence.value = -29939
        pack = test_sequence.pack()

        # garbage with our bytes in the middle
        test_bytes = b'\x1A\x3fadbsfkasdf;aslkfjasd;f' + pack + b'\x00\x00asdfn234r'
        for b in six.iterbytes(test_bytes):
            protocol_handler.feed(six.b(chr(b)))
        self.assertEqual(len(rx), 1)
        self.assertEqual(rx[0].value, -29939)


if __name__ == '__main__':
    unittest.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015 Digi International Inc. All Rights Reserved.
import sys
import socket

from six.moves import socketserver
from suitcase.protocol import StreamProtocolHandler
from suitcase.structure import Structure
from suitcase.fields import UBInt16, LengthField, Payload, UBInt8

FRAME_TYPE_ECHO_REQUEST = 0x00
FRAME_TYPE_ECHO_RESPONSE = 0x10


class EchoProtocolFrame(Structure):
    frame_type = UBInt8()
    payload_length = LengthField(UBInt16())
    payload = Payload(payload_length)


class EchoTCPHandler(socketserver.BaseRequestHandler):
    def _frame_received(self, request_frame):
        # frame is an instance of EchoProtocolFrame
        print("Received %r" % request_frame)
        if request_frame.frame_type == FRAME_TYPE_ECHO_REQUEST:
            response = EchoProtocolFrame()
            response.frame_type = FRAME_TYPE_ECHO_RESPONSE
            response.payload = "You sent %r" % request_frame.payload
            self.request.sendall(response.pack())
        else:
            print("Unexpected frame: %r" % request_frame)

    def setup(self):
        self.handler = StreamProtocolHandler(EchoProtocolFrame, self._frame_received)

    def handle(self):
        while True:
            self.handler.feed(self.request.recv(1024))


def client():
    def message_received(frame):
        print(frame)

    s = socket.socket()
    s.connect(('127.0.0.1', 7070))
    proto_handler = StreamProtocolHandler(EchoProtocolFrame, message_received)
    while True:
        input = raw_input("Data to send: ")
        request = EchoProtocolFrame()
        request.frame_type = FRAME_TYPE_ECHO_REQUEST
        request.payload = input
        s.sendall(request.pack())

        # to handle asynchronous messages, this would be done asynchronously
        proto_handler.feed(s.recv(1024))


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("Usage Client: %s -c" % sys.argv[0])
        print("Usage Server: %s -s" % sys.argv[1])
    if sys.argv[1] == '-c':
        client()
    elif sys.argv[1] == '-s':
        server = socketserver.TCPServer(('0.0.0.0', 7070), EchoTCPHandler)
        server.serve_forever()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015 Digi International Inc. All Rights Reserved.

import unittest

from suitcase.crc import crc16_ccitt, crc32, crc16_kermit


class TestCRC16CCITT(unittest.TestCase):
    def test_hello_world(self):
        crc = crc16_ccitt(b"Hello, world")
        self.assertEqual(crc, 0x3E99)


class TestCRC16Kermit(unittest.TestCase):
    def test_hello_world(self):
        crc = crc16_kermit(b"Hello, world")
        self.assertEqual(crc, 0xEB86)


class TestCRC32(unittest.TestCase):
    def test_hello_world(self):
        crc = crc32(b"Hello, world")
        self.assertEqual(crc, 0xE79AA9C2)


if __name__ == '__main__':
    unittest.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015 Digi International Inc. All Rights Reserved.

"""Example of UDP protocol datagram parsing with suitcase"""
import unittest

import six
from suitcase.fields import UBInt16, VariableRawPayload, LengthField, Magic, \
    UBInt8Sequence, DispatchField, DispatchTarget, UBInt8, UBInt32, BitField, BitNum, \
    BitBool
from suitcase.structure import Structure


class TCPFrameHeader(Structure):
    source_address = UBInt16()
    destination_address = UBInt16()
    sequence_number = UBInt32()
    acknowledgement_number = UBInt32()
    options = BitField(16,
                       data_offset=BitNum(4),
                       reserved=BitNum(3),
                       NS=BitBool(),
                       CWR=BitBool(),
                       ECE=BitBool(),
                       URG=BitBool(),
                       ACK=BitBool(),
                       PSH=BitBool(),
                       RST=BitBool(),
                       SYN=BitBool(),
                       FIN=BitBool()
                       )
    window_size = UBInt16()
    checksum = UBInt16()
    urgent_pointer = UBInt16()
    # TODO: additional options if data_offset > 5


class UDPFrame(Structure):
    source_port = UBInt16()
    destination_port = UBInt16()
    length = LengthField(UBInt16())
    checksum = UBInt16()
    data = VariableRawPayload(length)


class IPV4Frame(Structure):
    options = BitField(64,
                       version=BitNum(4),
                       internet_header_length=BitNum(4),
                       differentiated_services_code_point=BitNum(6),
                       explicit_congestion_notification=BitNum(2),
                       total_length=BitNum(16),
                       identification=BitNum(16),
                       flags=BitNum(3),
                       fragment_offset=BitNum(13),
                       )
    time_to_live = UBInt8()
    protocol = DispatchField(UBInt8())
    header_checksum = UBInt16()
    source_ip_address = UBInt32()
    destination_ip_address = UBInt32()


# This may not be fully accurate for ethernet as this is usually at
# a level well below anything Suitcase was designed to handle and there
# are some cases that we just don't handle (like a field being either
# a length of a type byte).
# 
# For instance, we don't allow for the option 802.1q vlan tagging
# which could come after mac_source
class EthernetFrame(Structure):
    preamble_sof = Magic('\xAA' * 8)  # 8 bytes of 10101010
    mac_dest = UBInt8Sequence(6)
    mac_source = UBInt8Sequence(6)
    ethertype = DispatchField(UBInt16())
    payload = DispatchTarget(None, ethertype, {
        0x0800: IPV4Frame,
        #         0x85DD: IPV6Message,
        #         0x0806: ARPMessage,
    })

    checksum = UBInt32()


class TestIPV4Frame(unittest.TestCase):
    def test(self):
        ipv4 = IPV4Frame()
        ipv4.options.version = 0x01
        ipv4.options.internet_header_length = 0
        ipv4.options.differentiated_services_code_point = 0x05
        ipv4.options.explicit_congestion_notification = 0x01
        ipv4.options.total_length = 0x03
        ipv4.options.identification = 0xFEFE
        ipv4.options.flags = 0x01
        ipv4.options.fragment_offset = 0x9F

        ipv4.source_ip_address = 0x91919191
        ipv4.destination_ip_address = 0x88776655

        ipv4.protocol = 0x06
        ipv4.time_to_live = 0x10
        ipv4.header_checksum = 0x90

        packed_value = six.b('\x10\x15\x00\x03\xfe\xfe \x9f\x10\x06\x00\x90\x91\x91\x91\x91\x88wfU')
        ipv4_2 = IPV4Frame()
        ipv4_2.unpack(packed_value)
        self.assertEqual(ipv4_2.options.version, 0x01)
        self.assertEqual(ipv4_2.options.internet_header_length, 0x00)


class TestUDPDatagram(unittest.TestCase):
    def test(self):
        dgram = UDPFrame()
        dgram.source_port = 9101
        dgram.destination_port = 9100
        dgram.checksum = 0x00
        dgram.data = six.b("Hello, UDP World!")
        packed = dgram.pack()
        self.assertEqual(packed,
                         six.b('#\x8d#\x8c\x00\x11\x00\x00Hello, UDP World!'))
        dgram2 = UDPFrame()
        dgram2.unpack(packed)
        self.assertEqual(dgram2.data, six.b("Hello, UDP World!"))


if __name__ == '__main__':
    unittest.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2019 Digi International Inc. All Rights Reserved.
import struct

import six
from suitcase.exceptions import SuitcaseChecksumException, SuitcaseProgrammingError, \
    SuitcaseParseError, SuitcaseException, SuitcasePackStructException
from six import BytesIO, StringIO


class FieldPlaceholder(object):
    """Internally used object that holds information about a field schema

    A FieldPlaceholder is what is actually instantiated and stored with
    the message schema class when a message schema is declared.  The
    placeholder stores all instantiation information needed to create
    the fields when the message object is instantiated

    """

    # record the number of instantiations of fields.  This is how
    # we can track the order of fields within a message
    _global_seqno = 0

    def __init__(self, cls, args, kwargs):
        self._field_seqno = FieldPlaceholder._global_seqno
        FieldPlaceholder._global_seqno += 1
        self.cls = cls
        self.args = args
        self.kwargs = kwargs
        self.__accessors = {}

    def create_instance(self, parent):
        """Create an instance based off this placeholder with some parent"""
        self.kwargs['instantiate'] = True
        self.kwargs['parent'] = parent
        instance = self.cls(*self.args, **self.kwargs)
        instance._field_seqno = self._field_seqno
        return instance

    def __getattr__(self, name):
        # In case the same field is being referenced more than once.
        if name in self.__accessors:
            return self.__accessors[name]
        accessor = FieldAccessor(self, name)
        self.__accessors[name] = accessor
        return accessor


class BaseField(object):
    """Base class for all Field instances

    Some magic is used whenever a field is created normally in order to
    allow for a declarative syntax without having to create explicit
    factory methods/classes everywhere.  By default (for a normal call)
    when a field is instantiated, it actually ends up returning a
    FieldPlaceholder object containing information about the field
    that has been declared.

    To get an actual instance of the field, the call to the construction
    must include a keyword argument ``instantiate`` that is set to some
    truthy value.  Typically, this instantiation logic is done by
    calling ``FieldPlaceholder.create_instance(parent)``.  This is the
    recommended way to construct a field as it ensure all invariants
    are in place (having a parent).

    :param instantiate: Create an actual instance instead of a placeholder
    :param parent: Specify the parent of this field (typically a Structure
        instance).

    """

    def __new__(cls, *args, **kwargs):
        instantiate = kwargs.pop('instantiate', False)
        if instantiate:
            return super(BaseField, cls).__new__(cls)
        else:
            return FieldPlaceholder(cls, args, kwargs)

    def __init__(self, *args, **kwargs):
        self._value = None
        self._parent = kwargs.get('parent')

    def _ph2f(self, placeholder):
        """Lookup a field given a field placeholder"""
        if issubclass(placeholder.cls, FieldAccessor):
            return placeholder.cls.access(self._parent, placeholder)
        return self._parent.lookup_field_by_placeholder(placeholder)

    def __repr__(self):
        return repr(self.getval())

    def is_substructure(self):
        """Return True if this field is (effectively) a SubstructureField"""
        return False

    def getval(self):
        return self._value

    def setval(self, value):
        self._value = value

    def unpack(self, data, **kwargs):  # type: (bytes, **bool) -> bytes
        """
        Given the raw data associated with a field, consume the data and return anything left over.
        :param data: Raw data to consume
        :returns: Remaining data
        """
        raise NotImplementedError("Field class must implement unpack.")

    @property
    def is_greedy(self):  # type: () -> bool
        """
        Should the parser consider this field to be greedy, i.e. that it wants to consume
        all remaining data in the input?

        This property currently is only meaningful on ``DispatchTarget``,
        where passing greedy=False allows for nested DispatchTarget structures.
        """
        return True


class CRCField(BaseField):
    r"""Field representing CRC (Cyclical Redundancy Check) in a message

    CRC checks and calculation frequently work quite differently
    from other fields in a protocol and as such are treated differently
    by the message container.  In particular, a CRCField requires
    special steps at either the beginning or end of the message
    pack/unpack process.  For each CRCField, we specify the following:

    :param field: The underlying field defining how the checksum will
        be stored.  This should match the size of the checksum in
        whatever algorithm is being used (16 bits for CRC16).
    :param algo: The algorithm to be used for performing the checksum.
       This is basically a function that takes a chunk of data and
       gives the checksum.  Several of these are provided out of the
       box in ``suitcase.crc``.
    :param start: The offset in the overall message at which we should
        start the checksum.  This may be positive (from the start) or
        negative (from the end of the message).
    :param end: The offset in the overall message at which we should
        end the checksum algo.  This may be positive (from the start)
        or negative (from the end of the message).

    A quick example of a message with a checksum is in order::

        class MyChecksummedMessage(Structure):
            soh = Magic('\x1f\x1f')
            message_id = UBInt16()
            sequence_number = UBInt8()
            payload_length = LengthField(UBInt16())
            payload = VariableRawPayload(payload_length)

            # crc starts after soh and ends before crc
            crc = CRCField(UBInt16(), crc16_ccitt, 2, -3)
            eof = Magic('~')

    """

    def __init__(self, field, algo, start, end, **kwargs):
        BaseField.__init__(self, **kwargs)
        self.field = field.create_instance(self._parent)
        self.field.setval(0)
        self.algo = algo
        self.start = start
        self.end = end
        self._value = None

    @property
    def bytes_required(self):
        return self.field.bytes_required

    def validate(self, data, offset):
        """Raises :class:`SuitcaseChecksumException` if not valid"""
        recorded_checksum = self.field.getval()

        # convert negative offset to positive
        if offset < 0:
            offset += len(data)

        # replace checksum region with zero
        data = b''.join((data[:offset],
                         b"\x00" * self.bytes_required,
                         data[offset + self.bytes_required:]))
        actual_checksum = self.algo(data[self.start:self.end])
        if recorded_checksum != actual_checksum:
            raise SuitcaseChecksumException(
                "recorded checksum %r did not match actual %r.  full data: %r",
                recorded_checksum, actual_checksum, data)

    def packed_checksum(self, data):
        """Given the data of the entire packet return the checksum bytes"""
        self.field.setval(self.algo(data[self.start:self.end]))
        sio = BytesIO()
        self.field.pack(sio)
        return sio.getvalue()

    def getval(self):
        return self.field.getval()

    def setval(self, *args):
        raise SuitcaseProgrammingError("CRC will be set automatically")

    def pack(self, stream):
        # write placeholder during the first pass
        stream.write(b'\x00' * self.field.bytes_required)

    def unpack(self, data, **kwargs):
        return self.field.unpack(data, **kwargs)


class Magic(BaseField):
    """Represent Byte Magic (fixed, expected sequence of bytes)"""

    def __init__(self, expected_sequence, **kwargs):
        BaseField.__init__(self, **kwargs)
        self.expected_sequence = expected_sequence
        self.bytes_required = len(self.expected_sequence)

    def getval(self):
        return self.expected_sequence

    def setval(self, *args):
        raise SuitcaseProgrammingError("One does not simply modify Magic")

    def pack(self, stream):
        stream.write(self.expected_sequence)

    def unpack(self, data, **kwargs):
        if not data == self.expected_sequence:
            raise SuitcaseParseError(
                "Expected sequence %r for magic field but got %r on "
                "message %r" % (self.expected_sequence, data, self._parent))
        return b''

    def __repr__(self):
        return "Magic(%r)" % (self.expected_sequence,)


class FieldProperty(BaseField):
    """Provide the ability to define "Property" getter/setters for other fields

    This is useful and preferable and preferable to inheritance if you want
    to provide a different interface for getting/setting one some field
    within a message.  Take the test case as an example::

        # define the message
        class MyMessage(Structure):
            _version = UBByteSequence(2)
            version = FieldProperty(_version,
                                    onget=lambda v: "%d.%02d" % (v[0], v[1]),
                                    onset=lambda v: tuple(int(x) for x
                                                          in v.split(".", 1)))

        msg = MyMessage()
        msg.unpack('\x10\x03')
        self.assertEqual(msg._version,(16, 3))
        self.assertEqual(msg.version, "16.03")

        msg.version = "22.7"
        self.assertEqual(msg._version, (22, 7))
        self.assertEqual(msg.version, "22.07")

    :param field: The field that is wrapped by this FieldProperty.  The value
        of this field is passed into or set by the associated get/set fns.
    :param onget: This is a function pointer to a function called to mutate
        the value returned on field access.  The function receives a single
        argument containing the value of the wrapped field.
    :oaram onset: This is a function pointer to a function called to map
        between the property and the underlying field.  The function takes
        a single parameter which is the value the property was set to.  It
        should return the value that the underlying field expects (or raise
        an exception if appropriate).

    """

    def __init__(self, field, onget=None, onset=None, **kwargs):
        BaseField.__init__(self, **kwargs)
        self.onget = onget
        self.onset = onset
        self.field = self._ph2f(field)
        self.bytes_required = 0

    def _default_onget(self, value):
        return value

    def _default_onset(self, value):
        return value

    def getval(self):
        onget = self.onget
        if onget is None:
            onget = self._default_onget
        value = self.field.getval()
        return onget(value)

    def setval(self, value):
        onset = self.onset
        if onset is None:
            onset = self._default_onset

        self.field.setval(onset(value))

    def unpack(self, data, **kwargs):
        return data

    def pack(self, stream):
        pass


class DispatchField(BaseField):
    """Decorate a field as a dispatch byte (used as conditional)

    A DispatchField is always used with a DispatchTarget within the same
    message (at some level).

    Example::

        class MyMessage(Structure):
            type = DispatchField(UBInt8())
            body = DispatchTarget(dispatch_field=type, dispatch_mapping={
                0x00: MessageType0,
                0x01: MessageType1,
            })

    :param field: The field containing the dispatch parameter.  This is
        typically an integer but could be any hashable object.

    """

    def __init__(self, field, **kwargs):
        BaseField.__init__(self, **kwargs)
        self.field = field.create_instance(self._parent)

    @property
    def bytes_required(self):
        return self.field.bytes_required

    def getval(self):
        return self.field.getval()

    def setval(self, value):
        return self.field.setval(value)

    def __repr__(self):
        return repr(self.field)

    def pack(self, stream):
        return self.field.pack(stream)

    def unpack(self, data, **kwargs):
        assert len(data) == self.bytes_required
        return self.field.unpack(data)


class DispatchTarget(BaseField):
    """Represent a conditional branch on some DispatchField

    Example::

        class MyMessage(Structure):
            type = DispatchField(UBInt8())
            body = DispatchTarget(dispatch_field=type, dispatch_mapping={
                0x00: MessageType0,
                0x01: MessageType1,
            })

    Observe that it is also possible to create a DispatchTarget based on
    a segment of a BitField::

        class MyMessage(Structure):
            version_and_type = BitField(8,
                version=BitNum(4),
                type=BitNum(4)
            )
            body = DispatchTarget(None, version_and_type.type, {
                0x0: MessageType0,
                0x1: MessageType1,
            })

    :param length_provider: The field providing a length value binding this
        message (if any).  Set this field to None to leave unconstrained.
    :param dispatch_field: The field being target for dispatch.  In most
        protocols this is an integer type byte or something similar.  The
        possible values for this field act as the keys for the dispatch.
    :param dispatch_mapping: This is a dictionary mapping dispatch_field
        values to associated message types to handle the remaining processing.
    :param greedy: If True (the default), the parser can assume that this field
        consumes all remaining data in the input. If False, the parser will only
        consume as much data as necessary to create the structure.
        This comes in handy when a DispatchTarget is inside of a substructure or another
        DispatchTarget, and is not the last field of the message.

    """

    def __init__(self, length_provider, dispatch_field,
                 dispatch_mapping, greedy=True, **kwargs):
        BaseField.__init__(self, **kwargs)
        if length_provider is None:
            self.length_provider = None
        else:
            self.length_provider = self._ph2f(length_provider)
            self.length_provider.associate_length_consumer(self)
        self.dispatch_field = self._ph2f(dispatch_field)
        self.dispatch_mapping = dispatch_mapping
        self.inverse_dispatch_mapping = dict((v, k) for (k, v)
                                             in dispatch_mapping.items())
        self.__greedy = greedy

    def _lookup_msg_type(self):
        target_key = self.dispatch_field.getval()
        target = self.dispatch_mapping.get(target_key, None)
        if target is None:
            target = self.dispatch_mapping.get(None, None)

        return target

    @property
    def is_greedy(self):
        return self.__greedy

    @property
    def bytes_required(self):
        if self.length_provider is None:
            return None
        else:
            return self.length_provider.get_adjusted_length()

    def getval(self):
        return self._value

    def setval(self, value):
        try:
            vtype = type(value)
            key = self.inverse_dispatch_mapping[vtype]
        except KeyError:
            raise SuitcaseProgrammingError("The type specified is not in the "
                                           "dispatch table")

        # OK, things check out.  Set both the value here and the
        # type byte value
        self._value = value
        value._parent = self._parent
        self.dispatch_field.setval(key)

    def pack(self, stream):
        return self._value._packer.write(stream)

    def unpack(self, data, **kwargs):
        target_msg_type = self._lookup_msg_type()
        if target_msg_type is None:
            target_msg_type = self.dispatch_mapping.get(None)
        if target_msg_type is None:  # still none
            raise SuitcaseParseError("Input data contains type byte not"
                                     " contained in mapping")
        message_instance = target_msg_type()
        self.setval(message_instance)
        return self._value.unpack(data, trailing=not self.is_greedy).read()


class LengthField(BaseField):
    """Wraps an existing field marking it as a LengthField

    This field wraps another field which is assumed to return an
    integer value.  A LengthField can be pointed to by a variable
    length field and the two will be coupled appropriately.

    :param length_field: The field providing the actual length value to
        be used.  This field should return an integer value representing
        the length (or a multiple of the length) as a return value
    :param get_length: If specified, this is a function which takes a reference
        to the encapsulated field and is responsible for reading the length
        out from that field.  By default, this just reads the value of the
        field (field is expected to be an integral value).
    :param set_length: If specified, this is a function which takes a reference
        to the encapsulated field and a length value.  Its responsibility
        is to move the provided length value into the field.  By default,
        the length value is just assigned to the field via `setval()`.
    :param multiplier: If specified, this multiplier is applied to the
        length.  If I specify a multiplier of 8, I am saying that each bit
        in the length field represents 8 bytes of actual payload length. By
        default the multiplier is 1 (1 bit/byte).

    """

    def __init__(self, length_field, get_length=None, set_length=None, multiplier=1, **kwargs):
        BaseField.__init__(self, **kwargs)
        self.multiplier = multiplier
        self.get_length = self._default_get_length if get_length is None else get_length
        self.set_length = self._default_set_length if set_length is None else set_length
        self.length_field = length_field.create_instance(self._parent)
        self.length_value_provider = None

    def __repr__(self):
        return repr(self.length_field)

    def _default_get_length(self, field):
        return field.getval()

    def _default_set_length(self, field, length):
        field._value = length  # TODO: use setval() [problem with DependentField tests]?

    def is_substructure(self):
        return self.length_field.is_substructure()

    @property
    def bytes_required(self):
        return self.length_field.bytes_required

    def getval(self):
        return self.length_field.getval()

    def setval(self, value):
        raise SuitcaseProgrammingError("Cannot set the value of a LengthField")

    def associate_length_consumer(self, target_field):
        def _length_value_provider():
            sio = BytesIO()
            target_field.pack(sio)
            target_field_length = len(sio.getvalue())
            if not target_field_length % self.multiplier == 0:
                raise SuitcaseProgrammingError("Payload length not divisible "
                                               "by %s" % self.multiplier)
            return (target_field_length // self.multiplier)

        self.length_value_provider = _length_value_provider

    def pack(self, stream):
        if self.length_value_provider is None:
            raise SuitcaseException("No length_provider added to this LengthField")
        self.set_length(self.length_field, self.length_value_provider())
        self.length_field.pack(stream)

    def unpack(self, data, **kwargs):
        assert len(data) == self.bytes_required
        return self.length_field.unpack(data, **kwargs)

    def get_adjusted_length(self):
        return self.get_length(self.length_field) * self.multiplier


class TypeField(BaseField):
    """Wraps an existing field marking it as a TypeField

    This field wraps another field which is assumed to return an
    integer value.  A TypeField can be pointed to by a variable
    length field and will fix that field's length.

    :param type_field: The field providing the type id that will
        be used to lookup a length value
    :param length_mapping: This is a dictionary mapping type_field
        values to associated length values.

    For example, a TypeField could be used as a replacement for a
    DispatchField so that the associated DispatchTarget is non-greedy,
    without the existence of a seperate LengthField. This enables
    a greedy field to follow the dispatch field, without the usage of
    complex ConditionalField setups::

        class Structure8Bit(Structure):
            value = UBInt8()

        class Structure16Bit(Structure):
            value = UBInt16()

        class DispatchStructure(Structure):
            type = TypeField(UBInt8(), {0x40: 1,
                                        0x80: 2})
            # Here the TypeField is providing both a size and type id
            dispatch = DispatchTarget(type, type, {0x40: Structure8Bit,
                                                   0x80: Structure16Bit})
            greedy = Payload()
    """

    def __init__(self, type_field, length_mapping, **kwargs):
        BaseField.__init__(self, **kwargs)
        self.type_field = type_field.create_instance(self._parent)
        self.length_mapping = length_mapping
        self.length_value_provider = None

    def _lookup_msg_length(self):
        target_key = self.type_field.getval()
        target_length = self.length_mapping.get(target_key, None)
        if target_length is None:
            target_length = self.length_mapping.get(None, None)

        return target_length

    def __repr__(self):
        return repr(self.type_field)

    def is_substructure(self):
        return self.type_field.is_substructure()

    @property
    def bytes_required(self):
        return self.type_field.bytes_required

    def getval(self):
        return self.type_field.getval()

    def setval(self, value):
        self.type_field.setval(value)

    def associate_length_consumer(self, target_field):
        def _length_value_provider():
            sio = BytesIO()
            target_field.pack(sio)
            target_field_length = len(sio.getvalue())
            if target_field_length != self.get_adjusted_length():
                raise SuitcaseProgrammingError("Payload length %i does not"
                                               " match length %i specified by type"
                                               % (target_field_length, self.get_adjusted_length()))
            return target_field_length

        self.length_value_provider = _length_value_provider

    def pack(self, stream):
        if self.length_value_provider is None:
            raise SuitcaseException("No length_provider added to this TypeField")
        # This will throw a SuitcasePackException if the length is not correct
        self.length_value_provider()

        self.type_field.pack(stream)

    def unpack(self, data, **kwargs):
        assert len(data) == self.bytes_required
        return self.type_field.unpack(data, **kwargs)

    def get_adjusted_length(self):
        return self._lookup_msg_length()


class ConditionalField(BaseField):
    """Field which may or may not be included depending on some condition

    In some protocols, there exist fields which may or may not be present
    depending on the values of other fields that would have already been
    parsed at this point in time.  Wrapping such fields in  a ConditionalField
    allows us to define a function to examine that state and only have the
    field capture the bytes if the conditions are right.

    :param field: The field which should handle the parsing if the condition
        evaluates to true.

    :param condition: This is a function which is given access to the parent
        message that is expected to return a boolean value.  If the value
        is true, then ``field`` will handle the bytes.  If not, then the
        field will be skipped and left with its default value (None).

    """

    def __init__(self, field, condition, **kwargs):
        BaseField.__init__(self, **kwargs)
        self.field = field.create_instance(self._parent)
        self.condition = condition

    def __repr__(self):
        if self.condition(self._parent):
            return repr(self.field)
        else:
            return "<ConditionalField: not included>"

    def is_substructure(self):
        return self.field.is_substructure()

    @property
    def bytes_required(self):
        if self.condition(self._parent):
            return self.field.bytes_required
        else:
            return 0

    def pack(self, stream):
        if self.condition(self._parent):
            self.field.pack(stream)

    def unpack(self, data, **kwargs):
        if self.condition(self._parent):
            return self.field.unpack(data, **kwargs)
        return data

    def getval(self):
        if not self.condition(self._parent):
            return None

        return self.field.getval()

    def setval(self, value):
        return self.field.setval(value)

    def associate_length_consumer(self, target_field):
        self.field.associate_length_consumer(target_field)

    def get_adjusted_length(self):
        return self.field.get_adjusted_length()


class Payload(BaseField):
    """Variable length raw (byte string) field

    This field is expected to be used with a LengthField.  The variable
    length field provides a value to be used by the LengthField on
    message pack and vice-versa for unpack.

    :param length_provider: The LengthField with which this variable
        length payload is associated.  If not included, it is assumed that
        the length_provider should consume the remainder of the bytes
        available in the string.  This is only valid in cases where the
        developer knows that they will be dealing with a fixed sequence
        of bytes (already boxed).

    """

    def __init__(self, length_provider=None, **kwargs):
        BaseField.__init__(self, **kwargs)
        if isinstance(length_provider, FieldPlaceholder):
            self.length_provider = self._ph2f(length_provider)
            self.length_provider.associate_length_consumer(self)
        else:
            self.length_provider = None

    @property
    def bytes_required(self):
        if self.length_provider is None:
            return None
        else:
            return self.length_provider.get_adjusted_length()

    def pack(self, stream):
        stream.write(self._value)

    def unpack(self, data, **kwargs):
        self._value = data
        assert self.bytes_required is None or len(data) == self.bytes_required
        return b''


# keep for backwards compatibility
VariableRawPayload = Payload


class BaseVariableByteSequence(BaseField):
    def __init__(self, make_format, length_provider, **kwargs):
        BaseField.__init__(self, **kwargs)
        self.make_format = make_format
        if length_provider is not None:
            self.length_provider = self._ph2f(length_provider)
            self.length_provider.associate_length_consumer(self)

    @property
    def bytes_required(self):
        return self.length_provider.get_adjusted_length()

    def pack(self, stream):
        sfmt = self.make_format(len(self._value))
        try:
            stream.write(struct.pack(sfmt, *self._value))
        except struct.error as e:
            raise SuitcasePackStructException(e)

    def unpack(self, data, **kwargs):
        assert len(data) == self.bytes_required

        length = self.bytes_required
        sfmt = self.make_format(length)
        try:
            self._value = struct.unpack(sfmt, data)
        except struct.error as e:
            raise SuitcasePackStructException(e)


class DependentField(BaseField):
    """Field populated by container packet at lower level

    It is sometimes the case that information from another layer of a
    messaging protocol be needed at another higher-level of the protocol.
    The DependentField is a way of declaring that a message at some layer
    is dependent on a field with some name from the parent layer.

    For instance, let's suppose that my protocol had an option byte that
    I wanted to include in some logic handling packets at some higher
    layer.  I could include that byte in my message as follows::

        class MyDependentMessage(Structure):
            ll_options = DependentField('proto_options')
            data = UBInt8Sequence(16)

        class LowerLevelProtocol(Structure):
            type = DispatchField(UBInt16())
            proto_options = UBInt8()
            body = DispatchTarget(None, type, {
                0x00: MyDependentMessage,
            })

    :param name: The name of the field from the parent message that we
        would like brought into our message namespace.

    """

    def __init__(self, name, **kwargs):
        BaseField.__init__(self, **kwargs)
        self.bytes_required = 0
        self.parent_field_name = name
        self.parent_field = None

    def _get_parent_field(self):
        if self.parent_field is None:
            message_parent = self._parent._parent
            target_field = message_parent.lookup_field_by_name(
                self.parent_field_name)
            self.parent_field = target_field
        return self.parent_field

    def __getattr__(self, attr):
        return getattr(self._get_parent_field(), attr)

    def pack(self, stream):
        pass

    def unpack(self, data, **kwargs):
        return data

    def getval(self):
        return self._get_parent_field().getval()

    def setval(self, value):
        return self._get_parent_field().setval(value)


class SubstructureField(BaseField):
    """Field which contains another non-greedy Structure.

    Often data-types are needed which cannot be easily
    described by a single field, but are representable as
    Structures. For example, Pascal style strings are prefixed
    with their length. It is often desirable to embed these
    data-types within another Structure, to avoid reimplementing
    them in every usage.


    A Pascal style string could be described as follows::

        from suitcase.structure import Structure
        from suitcase.fields import Payload, UBInt16, LengthField, SubstructureField

        class PascalString16(Structure):
            length = LengthField(UBInt16())
            value = Payload(length)

    A structure describing a name of a person might consist of two
    Pascal style strings. Instead of describing the ugly way::

        class NameUgly(Structure):
            first_length = LengthField(UBInt16())
            first_value = Payload(first_length)
            last_length = LengthField(UBInt16())
            last_value = Payload(last_length)

    it could be defined using a SubstructureField::

        class Name(Structure):
            first = SubstructureField(PascalString16)
            last = SubstructureField(PascalString16)
    """

    def __init__(self, substructure, **kwargs):
        BaseField.__init__(self, **kwargs)
        self.substructure = substructure
        self._value = substructure()
        self._value._parent = self._parent

    def is_substructure(self):
        return True

    @property
    def bytes_required(self):
        # We return None but do not count as a greedy field to the packer
        return None

    def pack(self, stream):
        stream.write(self._value.pack())

    def unpack(self, data, **kwargs):
        self._value = self.substructure()
        self._value._parent = self._parent
        return self._value.unpack(data, **kwargs).read()


class FieldArray(BaseField):
    """Field which contains a list of some other field.

    In some protocols, there exist repeated substructures which are present in a
    variable number. The variable nature of these fields make a DispatchField
    and DispatchTarget combination unsuitable; instead a FieldArray may be
    used to pack/unpack these fields to/from a list.

    :param substructure: The type of the array element. Must not be a greedy
        field, or else the array will only ever have one element containing
        the entire contents of the array.
    :param length_provider: The field providing a length value binding this
        message (if any).  Set this field to None to leave unconstrained.
    :param num_elements_provider: The field providing the number of elements
        in the array. Set this field to None to leave unconstrained.

    For example, one can imagine a message listing the zipcodes covered by
    a telephone area code. Depending on the population density, the number
    of zipcodes per area code could vary greatly. One implementation of this
    data structure could be::

        from suitcase.structure import Structure
        from suitcase.fields import UBInt16, FieldArray

        class ZipcodeStructure(Structure):
            zipcode = UBInt16()

        class TelephoneZipcodes(Structure):
            areacode = UBInt16()
            zipcodes = FieldArray(ZipcodeStructure)  # variable number of zipcodes

    """

    def __init__(self, substructure, length_provider=None, num_elements_provider=None, **kwargs):
        BaseField.__init__(self, **kwargs)
        self.substructure = substructure
        self._value = list()
        if isinstance(length_provider, FieldPlaceholder):
            self.length_provider = self._ph2f(length_provider)
            self.length_provider.associate_length_consumer(self)
        else:
            self.length_provider = None
        if isinstance(num_elements_provider, FieldPlaceholder):
            self.num_elements_provider = self._ph2f(num_elements_provider)
            self.num_elements_provider.length_value_provider = lambda: len(self._value)
        else:
            self.num_elements_provider = None


    @property
    def bytes_required(self):
        if self.length_provider is None:
            return None
        else:
            return self.length_provider.get_adjusted_length()

    @property
    def num_elements(self):
        if self.num_elements_provider is None:
            return None
        else:
            return self.num_elements_provider.get_adjusted_length()

    def pack(self, stream):
        for structure in self._value:
            stream.write(structure.pack())

    def unpack(self, data, **kwargs):
        length = self.bytes_required
        if length == 0 or (data == b"" and length is None) or self.num_elements == 0:
            # Array is empty.
            return data

        kwargs['trailing'] = True
        while True:
            structure = self.substructure()
            structure._parent = self._parent
            data = structure.unpack(data, **kwargs).read()
            self._value.append(structure)
            if data == b"":
                break
            if len(self._value) == self.num_elements:
                break

        if self.num_elements is not None and len(self._value) != self.num_elements:
            raise SuitcaseParseError("Expected %s elements but received %d." %
                                     (self.num_elements, len(self._value)))

        return data


class BaseFixedByteSequence(BaseField):
    """Base fixed-length byte sequence field"""

    def __init__(self, make_format, size, **kwargs):
        BaseField.__init__(self, **kwargs)
        self.bytes_required = size
        self.format = make_format(size)

    def pack(self, stream):
        try:
            stream.write(struct.pack(self.format, *self._value))
        except struct.error as e:
            raise SuitcasePackStructException(e)

    def unpack(self, data, **kwargs):
        try:
            self._value = struct.unpack(self.format, data)
            assert len(data) == self.bytes_required
            return b''
        except struct.error as e:
            raise SuitcasePackStructException(e)


def byte_sequence_factory_factory(make_format):
    def byte_sequence_factory(length_or_provider):
        if isinstance(length_or_provider, int):
            return BaseFixedByteSequence(make_format, length_or_provider)
        else:
            return BaseVariableByteSequence(make_format, length_or_provider)

    return byte_sequence_factory


UBInt8Sequence = byte_sequence_factory_factory(lambda l: ">" + "B" * l)
ULInt8Sequence = byte_sequence_factory_factory(lambda l: "<" + "B" * l)
SBInt8Sequence = byte_sequence_factory_factory(lambda l: ">" + "b" * l)
SLInt8Sequence = byte_sequence_factory_factory(lambda l: "<" + "b" * l)


class BaseStructField(BaseField):
    """Base for fields based very directly on python struct module formats

    It is expected that this class will be subclassed and customized by
    defining ``FORMAT`` at the class level.  ``FORMAT`` is expected to
    be a format string that could be used with struct.pack/unpack.  It
    should include endianness information.  If the ``FORMAT`` includes
    multiple elements, the default ``_unpack`` logic assumes that each
    element is a single byte and will OR these together.  To specialize
    this, _unpack should be overridden.

    """

    def __init__(self, **kwargs):
        BaseField.__init__(self, **kwargs)
        self._value = None
        self._keep_bytes = getattr(self, "KEEP_BYTES", None)
        if self._keep_bytes is not None:
            self.bytes_required = self._keep_bytes
        else:
            self.bytes_required = struct.calcsize(self.PACK_FORMAT)

    def pack(self, stream):
        try:
            keep_bytes = getattr(self, 'KEEP_BYTES', None)
            if keep_bytes is not None:
                if self.PACK_FORMAT[0] == b">"[0]:  # The element access makes this compatible with Python 2 and 3
                    to_write = struct.pack(self.PACK_FORMAT, self._value)[-keep_bytes:]
                else:
                    to_write = struct.pack(self.PACK_FORMAT, self._value)[:keep_bytes]
            else:
                to_write = struct.pack(self.PACK_FORMAT, self._value)
        except struct.error as e:
            raise SuitcasePackStructException(e)
        stream.write(to_write)

    def unpack(self, data, **kwargs):
        value = 0
        if self.UNPACK_FORMAT[0] == b">"[0]:  # The element access makes this compatible with Python 2 and 3
            for i, byte in enumerate(reversed(struct.unpack(self.UNPACK_FORMAT, data))):
                value |= (byte << (i * 8))
        else:
            for i, byte in enumerate(struct.unpack(self.UNPACK_FORMAT, data)):
                value |= (byte << (i * 8))
        self._value = value
        assert len(data) == self.bytes_required
        return b''


# ==============================================================================
# Unsigned Big Endian
# ==============================================================================
class UBInt8(BaseStructField):
    """Unsigned Big Endian 8-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b">B"


class UBInt16(BaseStructField):
    """Unsigned Big Endian 16-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b">H"


class UBInt24(BaseStructField):
    """Unsigned Big Endian 24-bit integer field"""
    KEEP_BYTES = 3
    PACK_FORMAT = b">I"
    UNPACK_FORMAT = b">BBB"


class UBInt32(BaseStructField):
    """Unsigned Big Endian 32-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b">I"


class UBInt40(BaseStructField):
    """Unsigned Big Endian 40-bit integer field"""
    KEEP_BYTES = 5
    PACK_FORMAT = b">Q"
    UNPACK_FORMAT = b">BBBBB"


class UBInt48(BaseStructField):
    """Unsigned Big Endian 48-bit integer field"""
    KEEP_BYTES = 6
    PACK_FORMAT = b">Q"
    UNPACK_FORMAT = b">BBBBBB"


class UBInt56(BaseStructField):
    """Unsigned Big Endian 56-bit integer field"""
    KEEP_BYTES = 7
    PACK_FORMAT = b">Q"
    UNPACK_FORMAT = b">BBBBBBB"


class UBInt64(BaseStructField):
    """Unsigned Big Endian 64-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b">Q"


# ==============================================================================
# Signed Big Endian
# ==============================================================================
class SBInt8(BaseStructField):
    """Signed Big Endian 8-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b">b"


class SBInt16(BaseStructField):
    """Signed Big Endian 16-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b">h"


class SBInt24(BaseStructField):
    """Signed Big Endian 24-bit integer field"""
    KEEP_BYTES = 3
    PACK_FORMAT = b">i"
    UNPACK_FORMAT = b">bBB"


class SBInt32(BaseStructField):
    """Signed Big Endian 32-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b">i"


class SBInt40(BaseStructField):
    """Signed Big Endian 40-bit integer field"""
    KEEP_BYTES = 5
    PACK_FORMAT = b">q"
    UNPACK_FORMAT = b">bBBBB"


class SBInt48(BaseStructField):
    """Signed Big Endian 48-bit integer field"""
    KEEP_BYTES = 6
    PACK_FORMAT = b">q"
    UNPACK_FORMAT = b">bBBBBB"


class SBInt56(BaseStructField):
    """Signed Big Endian 56-bit integer field"""
    KEEP_BYTES = 7
    PACK_FORMAT = b">q"
    UNPACK_FORMAT = b">bBBBBBB"


class SBInt64(BaseStructField):
    """Signed Big Endian 64-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b">q"


class SBFloat32(BaseStructField):
    """Unsigned Big Endian 32-bit float field"""
    PACK_FORMAT = UNPACK_FORMAT = b">f"


class SBFloat64(BaseStructField):
    """Unsigned Big Endian 64-bit float field"""
    PACK_FORMAT = UNPACK_FORMAT = b">d"


# ==============================================================================
# Unsigned Little Endian
# ==============================================================================
class ULInt8(BaseStructField):
    """Unsigned Little Endian 8-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b"<B"


class ULInt16(BaseStructField):
    """Unsigned Little Endian 16-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b"<H"


class ULInt24(BaseStructField):
    """Unsigned Little Endian 24-bit integer field"""
    KEEP_BYTES = 3
    PACK_FORMAT = b"<I"
    UNPACK_FORMAT = b"<BBB"


class ULInt32(BaseStructField):
    """Unsigned Little Endian 32-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b"<I"


class ULInt40(BaseStructField):
    """Unsigned Little Endian 40-bit integer field"""
    KEEP_BYTES = 5
    PACK_FORMAT = b"<Q"
    UNPACK_FORMAT = b"<BBBBB"


class ULInt48(BaseStructField):
    """Unsigned Little Endian 48-bit integer field"""
    KEEP_BYTES = 6
    PACK_FORMAT = b"<Q"
    UNPACK_FORMAT = b"<BBBBBB"


class ULInt56(BaseStructField):
    """Unsigned Little Endian 56-bit integer field"""
    KEEP_BYTES = 7
    PACK_FORMAT = b"<Q"
    UNPACK_FORMAT = b"<BBBBBBB"


class ULInt64(BaseStructField):
    """Unsigned Little Endian 64-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b"<Q"


# ==============================================================================
# Signed Little Endian
# ==============================================================================
class SLInt8(BaseStructField):
    """Signed Little Endian 8-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b"<b"


class SLInt16(BaseStructField):
    """Signed Little Endian 16-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b"<h"


class SLInt24(BaseStructField):
    """Signed Little Endian 24-bit integer field"""
    KEEP_BYTES = 3
    PACK_FORMAT = b"<i"
    UNPACK_FORMAT = b"<BBb"


class SLInt32(BaseStructField):
    """Signed Little Endian 32-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b"<i"


class SLInt40(BaseStructField):
    """Signed Little Endian 40-bit integer field"""
    KEEP_BYTES = 5
    PACK_FORMAT = b"<q"
    UNPACK_FORMAT = b"<BBBBb"


class SLInt48(BaseStructField):
    """Signed Little Endian 48-bit integer field"""
    KEEP_BYTES = 6
    PACK_FORMAT = b"<q"
    UNPACK_FORMAT = b"<BBBBBb"


class SLInt56(BaseStructField):
    """Signed Little Endian 56-bit integer field"""
    KEEP_BYTES = 7
    PACK_FORMAT = b"<q"
    UNPACK_FORMAT = b"<BBBBBBb"


class SLInt64(BaseStructField):
    """Signed Little Endian 64-bit integer field"""
    PACK_FORMAT = UNPACK_FORMAT = b"<q"

    
class SLFloat32(BaseStructField):
    """Unsigned Little Endian 32-bit float field"""
    PACK_FORMAT = UNPACK_FORMAT = b"<f"


class SLFloat64(BaseStructField):
    """Unsigned Little Endian 64-bit float field"""
    PACK_FORMAT = UNPACK_FORMAT = b"<d"


# ==============================================================================
# BitField and Bits
# ==============================================================================
def bitfield_placeholder_factory_factory(cls):
    def _factory_fn(*args, **kwargs):
        return cls(*args, **kwargs)

    return _factory_fn


class _BitFieldFieldPlaceholder(object):
    _global_seqno = 0

    def __init__(self, cls, args, kwargs):
        self.cls = cls
        self.args = args
        self.kwargs = kwargs
        self.sequence_number = _BitFieldFieldPlaceholder._global_seqno
        _BitFieldFieldPlaceholder._global_seqno += 1

    def create_instance(self):
        kwargs = self.kwargs
        kwargs['instantiate'] = True
        return self.cls(*self.args, **self.kwargs)


class _BitFieldField(object):
    def __init__(self, *args, **kwargs):
        pass

    def __repr__(self):
        return repr(self.viewget())

    def __new__(cls, *args, **kwargs):
        if 'instantiate' in kwargs:
            return super(_BitFieldField, cls).__new__(cls)
        else:
            return _BitFieldFieldPlaceholder(cls, args, kwargs)


class _BitBool(_BitFieldField):
    def __init__(self, **kwargs):
        _BitFieldField.__init__(self, **kwargs)
        self.size = 1
        self._value = 0

    def getval(self):
        return self._value

    def setval(self, value):
        self._value = value

    def viewget(self):
        return (self._value == 1)

    def viewset(self, value):
        if value:
            self._value = 1
        else:
            self._value = 0


class _BitNum(_BitFieldField):
    def __init__(self, size, **kwargs):
        _BitFieldField.__init__(self, **kwargs)
        self.size = size
        self._value = 0

    def getval(self):
        return self._value

    viewget = getval

    def setval(self, value):
        self._value = value

    viewset = setval


BitNum = bitfield_placeholder_factory_factory(_BitNum)
BitBool = bitfield_placeholder_factory_factory(_BitBool)


class BitField(BaseField):
    """Represent a sequence of bytes broken down into bit segments

    Bit segments may be any BitFieldField instance, with the two most
    commonly used fields being:

    * BitBool(): A single bit flag that is either True or False
    * BitNum(bits): A multi-bit field treated like a big-endian integral
           value.

    Example Usage::

        class TCPFrameHeader(Structure):
            source_address = UBInt16()
            destination_address = UBInt16()
            sequence_number = UBInt32()
            acknowledgement_number = UBInt32()
            options = BitField(16,
                data_offset=BitNum(4),
                reserved=BitNum(3),
                NS=BitBool(),
                CWR=BitBool(),
                ECE=BitBool(),
                URG=BitBool(),
                ACK=BitBool(),
                PSH=BitBool(),
                RST=BitBool(),
                SYN=BitBool(),
                FIN=BitBool()
            )
            window_size = UBInt16()
            checksum = UBInt16()
            urgent_pointer = UBInt16()

        tcp_packet = TCPFrameHeader()
        o = tcp_packet.options
        o.data_offset = 3
        o.NS = True
        o.CWR = False
        o.
        # ... so on, so forth

    """

    def __init__(self, number_bits, field=None, **kwargs):
        BaseField.__init__(self, **kwargs)
        self._ordered_bitfields = []
        self._bitfield_map = {}
        if number_bits % 8 != 0:
            raise SuitcaseProgrammingError("Number of bits must be a factor of "
                                           "8, was %d" % number_bits)

        self.number_bits = number_bits
        self.number_bytes = number_bits // 8
        self.bytes_required = self.number_bytes
        if field is None:
            field = {
                1: UBInt8,
                2: UBInt16,
                3: UBInt24,
                4: UBInt32,
                5: UBInt40,
                6: UBInt48,
                7: UBInt56,
                8: UBInt64,
            }[self.number_bytes]()
        self._field = field.create_instance(self._parent)

        placeholders = []
        for key, value in six.iteritems(kwargs):
            if isinstance(value, _BitFieldFieldPlaceholder):
                placeholders.append((key, value))

        for key, placeholder in sorted(placeholders, key=lambda kv: kv[1].sequence_number):
            value = placeholder.create_instance()
            self._bitfield_map[key] = value
            self._ordered_bitfields.append((key, value))

    def __getattr__(self, key):
        if key in self.__dict__.get('_bitfield_map', {}):
            return self._bitfield_map[key].viewget()
        return object.__getattribute__(self, key)

    def __setattr__(self, key, value):
        if key in self.__dict__.get('_bitfield_map', {}):
            self._bitfield_map[key].viewset(value)
        else:
            self.__dict__[key] = value

    def __repr__(self):
        sio = StringIO()
        sio.write("BitField(\n")
        for key, field in self._ordered_bitfields:
            sio.write("    %s=%r,\n" % (key, field))
        sio.write("  )")
        return sio.getvalue()

    def getval(self):
        return self

    def setval(self, value):
        raise SuitcaseProgrammingError("Setting the value of a bitfield "
                                       "directly is prohibited")

    def pack(self, stream):
        value = 0
        shift = self.number_bits
        for _key, field in self._ordered_bitfields:
            shift -= field.size
            mask = (2 ** field.size - 1)  # mask off size bits
            value |= ((field.getval() & mask) << shift)

        self._field.setval(value)
        sio = BytesIO()
        self._field.pack(sio)
        out = sio.getvalue()[-self.number_bytes:]
        stream.write(out)

    def unpack(self, data, **kwargs):
        self._field.unpack(data, **kwargs)
        value = self._field.getval()
        shift = self.number_bits
        for _key, field in self._ordered_bitfields:
            shift -= field.size
            mask = (2 ** field.size - 1)
            fval = (value >> shift) & mask
            field.setval(fval)


class FieldAccessor(BaseField):
    """Represent attribute access to another field in the structure

    Not meant to be created directly by the user code.
    Instead, this is created automatically when a pattern like the following
    is used::

        class S(Structure):
            bits = BitField(8, top=BitNum(4), bottom=BitNum(4))
            body = DispatchTarget(None, bits.top, { ... })

    `bits.top` is a FieldAccessor, and is responsible for providing the
    value of the `top` field of `bits` when asked for, and updating that
    field when set.

    """

    def __init__(self, field, name, **kwargs):
        BaseField.__init__(self, field, name, **kwargs)
        self._field = field
        self._name = name
        self.bytes_required = 0

    @classmethod
    def access(cls, parent, placeholder):
        """Resolve the deferred field attribute access.

        :param cls: the FieldAccessor class
        :param parent: owning structure of the field being accessed
        :param placeholder: FieldPlaceholder object which holds our info
        :returns: FieldAccessor instance for that field's attribute
        """
        try:
            return parent.lookup_field_by_placeholder(placeholder)
        except KeyError:
            field_placeholder, name = placeholder.args
            # Find the field whose attribute is being accessed.
            field = parent.lookup_field_by_placeholder(field_placeholder)
            # Instantiate the real FieldAccessor that wraps the attribute.
            accessor = cls(field, name, parent=parent, instantiate=True)
            # Keep this instance alive, and reuse it for future references.
            parent._placeholder_to_field[placeholder] = accessor
            return accessor

    def getval(self):
        return getattr(self._field, self._name)

    def setval(self, value):
        setattr(self._field, self._name, value)

    def pack(self, stream):  # pragma: no cover
        pass

    def unpack(self, data, **kwargs):  # pragma: no cover
        pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2019 Digi International Inc. All Rights Reserved.
import sys
import os
import six
from suitcase.exceptions import SuitcaseException, \
    SuitcasePackException, SuitcaseParseError
from suitcase.fields import FieldArray, FieldPlaceholder, CRCField, SubstructureField, \
    ConditionalField, FieldAccessor, FieldProperty
from six import BytesIO


class ParseError(Exception):
    """Exception raised when there is an error parsing"""


class Packer(object):
    """Object responsible for packing/unpacking bytes into/from fields"""

    def __init__(self, ordered_fields, crc_field):
        self.crc_field = crc_field
        self.ordered_fields = ordered_fields

    def pack(self):
        # type: () -> bytes
        sio = BytesIO()
        self.write(sio)
        return sio.getvalue()

    def write(self, stream):
        # type: (BytesIO) -> None
        # now, pack everything in
        crc_fields = []
        for name, field in self.ordered_fields:
            try:
                if isinstance(field, CRCField):
                    crc_offset = stream.tell()
                    field.pack(stream)
                    crc_fields.append((field, crc_offset))
                else:
                    field.pack(stream)
            except SuitcaseException:
                raise  # just reraise the same exception object
            except Exception:
                # keep the original traceback information, see
                # http://stackoverflow.com/questions/3847503/wrapping-exceptions-in-python
                exc_type = SuitcasePackException
                _, exc_value, exc_traceback = sys.exc_info()
                exc_value = exc_type("Unexpected exception during pack of %r: %s" % (name, str(exc_value)))
                six.reraise(exc_type, exc_value, exc_traceback)

        # if there is a crc value, seek back to the field and
        # pack it with the right value
        if len(crc_fields) > 0:
            data = stream.getvalue()
            for field, offset in crc_fields:
                stream.seek(offset)
                checksum_data = self.crc_field.packed_checksum(data)
                stream.write(checksum_data)

    def unpack(self, data, trailing=False):
        # type: (bytes, bool) -> BytesIO
        stream = BytesIO(data)
        self.unpack_stream(stream)
        stream.tell()
        if trailing:
            return stream
        elif stream.tell() != len(data):
            raise SuitcaseParseError("Structure fully parsed but additional bytes remained.  Parsing "
                                     "consumed %d of %d bytes" %
                                     (stream.tell(), len(data)))
        return stream

    def unpack_stream(self, stream):
        # type: (BytesIO) -> None
        """Unpack bytes from a stream of data field-by-field

        In the most basic case, the basic algorithm here is as follows::

            for _name, field in self.ordered_fields:
               length = field.bytes_required
               data = stream.read(length)
               field.unpack(data)

        This logic is complicated somewhat by the handling of variable length
        greedy fields (there may only be one).  The logic when we see a
        greedy field (bytes_required returns None) in the stream is to
        pivot and parse the remaining fields starting from the last and
        moving through the stream backwards.  There is also some special
        logic present for dealing with checksum fields.

        """
        crc_fields = []
        greedy_field = None
        greedy_field_name = None
        # go through the fields from first to last.  If we hit a greedy
        # field, break out of the loop
        for i, (name, field) in enumerate(self.ordered_fields):
            if isinstance(field, CRCField):
                crc_fields.append((field, stream.tell()))
            length = field.bytes_required
            if field.is_substructure():
                remaining_data = stream.getvalue()[stream.tell():]
                after_unpack = field.unpack(remaining_data, trailing=True)
                consumed = len(remaining_data) - len(after_unpack)
                # We need to fast forward by as much as was consumed by the structure
                stream.seek(stream.tell() + consumed)
                continue
            elif length is None and field.is_greedy:
                # If length is None, this is assumed to be a greedy field.
                # But we check is_greedy so that non-greedy fields are supported.

                if isinstance(field, FieldArray) and field.num_elements is not None:
                    # Read the data greedily now, and we'll backtrack after enough elements have been read.
                    data = stream.read()
                else:
                    greedy_field = field
                    greedy_field_name = name
                    break
            else:
                data = stream.read(length)
                if length is not None and len(data) != length:
                    raise SuitcaseParseError("While attempting to parse field "
                                             "%r we tried to read %s bytes but "
                                             "we were only able to read %s." %
                                             (name, length, len(data)))

            try:
                unused_data = field.unpack(data)
                stream.seek(-len(unused_data or b""), os.SEEK_CUR)
            except SuitcaseException:
                raise  # just re-raise these
            except Exception:
                exc_type = SuitcaseParseError
                _, exc_value, exc_traceback = sys.exc_info()
                exc_value = exc_type("Unexpected exception while unpacking field %r: %s" % (name, str(exc_value)))
                six.reraise(exc_type, exc_value, exc_traceback)

        if greedy_field is not None:
            remaining_data = stream.read()
            inverted_stream = BytesIO(remaining_data[::-1])

            # work through the remaining fields in reverse order in order
            # to narrow in on the right bytes for the greedy field
            reversed_remaining_fields = self.ordered_fields[(i + 1):][::-1]
            for _name, field in reversed_remaining_fields:
                if isinstance(field, CRCField):
                    crc_fields.append(
                        (field, -inverted_stream.tell() - field.bytes_required))
                length = field.bytes_required
                if length is None and field.is_greedy:
                    raise SuitcaseParseError(
                        "While attempting to parse greedy field %r we found "
                        "another greedy field, %r. There can only be one greedy"
                        "field." %
                        (greedy_field_name, _name)
                    )

                data = inverted_stream.read(length)[::-1]
                if len(data) != length:
                    raise SuitcaseParseError("While attempting to parse field "
                                             "%r we tried to read %s bytes but "
                                             "we were only able to read %s." %
                                             (_name, length, len(data)))
                try:
                    field.unpack(data)
                except SuitcaseException:
                    raise  # just re-raise these
                except Exception:
                    exc_type = SuitcaseParseError
                    _, exc_value, exc_traceback = sys.exc_info()
                    exc_value = exc_type("Unexpected exception while unpacking field %r: %s" % (name, str(exc_value)))
                    six.reraise(exc_type, exc_value, exc_traceback)

            greedy_data_chunk = inverted_stream.read()[::-1]
            greedy_field.unpack(greedy_data_chunk)

        if crc_fields:
            data = stream.getvalue()
            for (crc_field, offset) in crc_fields:
                crc_field.validate(data, offset)


class StructureMeta(type):
    """Metaclass for all structure objects

    When a class with this metaclass is created, we look for any
    FieldProperty instances associated with the class and record
    those for use later on.

    """

    def __new__(cls, name, bases, dct):
        #  find all the placeholders in this class declaration and store
        # them away.  Add name mangling to the original fields so they
        # do not get in the way.
        dct['_field_placeholders'] = {}
        dct['_crc_field'] = None
        for key, value in list(dct.items()):  # use a copy, we mutate dct
            if isinstance(value, FieldPlaceholder):
                if issubclass(value.cls, FieldAccessor):
                    # Wrap the accessor in a simple FieldProperty,
                    # so that the following usage model is supported:
                    #
                    #     class S(Structure):
                    #         bits = BitField(...)
                    #         segment = bits.segment
                    value = FieldProperty(value)

                dct['_field_placeholders'][key] = value
                dct['__%s' % key] = value
                del dct[key]
                if value.cls == CRCField:
                    dct['_crc_field'] = value

        sorted_fields = list(sorted(dct['_field_placeholders'].items(),
                                    key=lambda kv: kv[1]._field_seqno))
        dct['_sorted_fields'] = sorted_fields
        return type.__new__(cls, name, bases, dct)


@six.add_metaclass(StructureMeta)
class Structure(object):
    r"""Base class for message schema declaration

    ``Structure`` forms the core of the Suitcase library and allows for
    a declarative syntax for specifying packet schemas and associated
    methods for transforming these schemas into packed bytes (and vice-versa).

    Here's an example showing how one might specify the format for a UDP
    Datagram::


        >>> from suitcase.fields import UBInt16, LengthField, VariableRawPayload
        >>> class UDPDatagram(Structure):
        ...     source_port = UBInt16()
        ...     destination_port = UBInt16()
        ...     length = LengthField(UBInt16())
        ...     checksum = UBInt16()
        ...     data = VariableRawPayload(length)

    From this we have a near-ideal form for packing and parsing packet
    data following the schema::

        >>> def printb(s):
        ...     print(repr(s).replace("b'", "'").replace("u'", "'"))
        ...
        >>> dgram = UDPDatagram()
        >>> dgram.source_port = 9110
        >>> dgram.destination_port = 1001
        >>> dgram.checksum = 27193
        >>> dgram.data = b"Hello, world!"
        >>> printb(dgram.pack())
        '#\x96\x03\xe9\x00\rj9Hello, world!'
        >>> dgram2 = UDPDatagram()
        >>> dgram2.unpack(dgram.pack())
        >>> dgram2
        UDPDatagram (
          source_port=9110,
          destination_port=1001,
          length=13,
          checksum=27193,
          data=...'Hello, world!',
        )

    Initialization via keyword argument is also supported::

        >>> dgram = UDPDatagram(source_port=9110,
        ...                     destination_port=1001,
        ...                     checksum=27193,
        ...                     data=b"Hello, world!")
        ...
        >>> printb(dgram.pack())
        '#\x96\x03\xe9\x00\rj9Hello, world!'
        >>> dgram2 = UDPDatagram()
        >>> dgram2.unpack(dgram.pack())
        >>> dgram2
        UDPDatagram (
          source_port=9110,
          destination_port=1001,
          length=13,
          checksum=27193,
          data=...'Hello, world!',
        )

    """

    @classmethod
    def from_data(cls, data):
        """Create a new, populated message from some data

        This factory method is identical to doing the following, it just takes
        one line instead of two and looks nicer in general::

            m = MyMessage()
            m.unpack(data)

        Can be rewritten as just::

            m = MyMessage.from_data(data)

        """
        m = cls()
        m.unpack(data)
        return m

    def __init__(self, **kwargs):
        self._key_to_field = {}
        self._parent = None
        self._sorted_fields = []
        self._placeholder_to_field = {}
        if self.__class__._crc_field is None:
            self._crc_field = None
        else:
            self._crc_field = self.__class__._crc_field.create_instance(self)
        for key, field_placeholder in self.__class__._sorted_fields:
            field = field_placeholder.create_instance(self)
            self._key_to_field[key] = field
            self._placeholder_to_field[field_placeholder] = field
            self._sorted_fields.append((key, field))
        self._packer = Packer(self._sorted_fields, self._crc_field)
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getattr__(self, key):
        k2f = self.__dict__.get('_key_to_field', {})
        if key in k2f:
            field = self._key_to_field[key]
            return field.getval()
        raise AttributeError

    def __setattr__(self, key, value):
        k2f = self.__dict__.get('_key_to_field', {})
        if key in k2f:
            field = self._key_to_field[key]
            return field.setval(value)
        return object.__setattr__(self, key, value)

    def __dir__(self):
        return dir(type(self)) + [str(k) for k, v in self._sorted_fields]

    def __iter__(self):
        return iter(self._sorted_fields)

    def __repr__(self):
        output = "%s (\n" % self.__class__.__name__
        for field_name, field in self:
            output += "  %s=%s,\n" % (field_name, field)
        output += ")"
        return output

    def lookup_field_by_name(self, name):
        for fname, field in self:
            if name == fname:
                return field
        raise KeyError

    def lookup_field_by_placeholder(self, placeholder):
        return self._placeholder_to_field[placeholder]

    def unpack(self, data, trailing=False):
        # type: (bytes, bool) -> BytesIO
        # If we asked to unpack while leaving any trailing bytes,
        # make sure to specify it's okay for there to be trailing bytes.
        return self._packer.unpack(data, trailing=trailing)

    def pack(self):
        # type: () -> bytes
        return self._packer.pack()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015 Digi International Inc. All Rights Reserved.
import binascii
from suitcase.fields import Payload
from suitcase.fields import UBInt16
from suitcase.fields import BitNum
from suitcase.fields import BitBool
from suitcase.fields import BitField
from suitcase.structure import Structure

class DNSMessage(Structure):
    identification = UBInt16()
    fields = BitField(16,
                      is_reply=BitBool(),  # QR
                      opcode=BitNum(4),
                      truncated=BitBool(),
                      recursion_desired=BitBool(),
                      ra=BitBool(),
                      z=BitBool(),
                      non_authenticated_data_acceptable=BitBool(),
                      cd=BitBool(),
                      rcode=BitNum(4),
                  )
    total_questions = UBInt16()
    total_answers_rrs = UBInt16()
    total_authority_rrs = UBInt16()
    total_additional_rrs = UBInt16()
    data = Payload()  # greedy





#
# Parsing Example (Line 40)
#
dns_request_hex = (
    b"28b30100000100000000000004646f63730464696769"
    b"03636f6d0000010001")
dns_request_data = binascii.unhexlify(dns_request_hex)
print(DNSMessage.from_data(dns_request_data))


#
# Packing Example (Line 50)
#
m = DNSMessage()
m.identification = 0x1234
m.fields.is_reply = False
m.fields.opcode = 0
m.fields.truncated = False
m.fields.recursion_desired = True
m.fields.ra = False
m.fields.z = False
m.fields.non_authenticated_data_acceptable = False
m.fields.cd = False
m.fields.rcode = 0
m.total_questions = 1
m.total_answers_rrs = 0
m.total_authority_rrs = 0
m.total_additional_rrs = 0
m.data = b"docs.digi.com\x00\x00\x00\x00"

print(repr(m.pack()))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015 Digi International Inc. All Rights Reserved.

"""Define an exception hierarchy used within Suitcase

The basic rule is that users of the Suitcase library should always
be able to catch the base :class:`.SuitcaseException` and be confident
that they will not get some other random exception.  Catching
more specific subclasses of this exception should then be possible.

There are still some cases (especially when constructing a field) that
base exceptions (like TypeError, ValueError, or TypeError) might still be
raised.  In general, however, any case where things are dynamic, we do
our best to raise a :class:`.SuitcaseException` with meaningful information.
"""


class SuitcaseException(Exception):
    """Base exception acting as the root for other exceptions"""


class SuitcaseChecksumException(SuitcaseException):
    """Exception raised when a checksum does not match"""


class SuitcaseProgrammingError(SuitcaseException):
    """Exception raised when somebody is doing something wrong"""


class SuitcaseParseError(SuitcaseException):
    """Raised when there is a problem parsing bytes into a suitcase schema"""


class SuitcasePackException(SuitcaseException):
    """Raised when there is an error packing a message"""


class SuitcasePackStructException(SuitcasePackException):
    """Raised on an error packing a struct field"""

    def __init__(self, struct_exception):
        self.struct_exception = struct_exception

    def __repr__(self):
        return repr(self.struct_exception)

    def __str__(self):
        return str(self.struct_exception)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015 Digi International Inc. All Rights Reserved.
import os
from setuptools import setup, find_packages


def get_long_description():
    long_description = open('README.md', 'r').read()
    try:
        import subprocess
        import pandoc

        process = subprocess.Popen(
            ['which pandoc'],
            shell=True,
            stdout=subprocess.PIPE,
            universal_newlines=True)

        pandoc_path = process.communicate()[0]
        pandoc_path = pandoc_path.strip('\n')

        pandoc.core.PANDOC_PATH = pandoc_path

        doc = pandoc.Document()
        doc.markdown = long_description.encode('utf-8')
        long_description = doc.rst
        open("README.rst", "wb").write(long_description)
    except:
        import traceback; traceback.print_exc()
        if os.path.exists("README.rst"):
            long_description = open("README.rst").read()
            if not long_description:
                raise RuntimeError("Failed to generate README.rst")
        else:
            print("Could not find pandoc or convert properly")
            print("  make sure you have pandoc (system) and pyandoc (python module) installed")
            raise RuntimeError("hmm")

    return long_description


setup(
    name='suitcase',
    version='0.12.0',
    url="https://github.com/digidotcom/python-suitcase",
    description='A library for specifying/parsing/packing binary protocols',
    long_description=get_long_description(),
    author="Paul Osborne",
    author_email="paul.osborne@digi.com",
    packages=find_packages(),
    keywords = ("suitcase, construct, declarative, dsl, protocol,"
                "binary, parser, builder, pack, unpack"),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries",
    ],
    install_requires=[
        "six",
    ],
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015 Digi International Inc. All Rights Reserved.

"""Define protocol handlers for different classes of protocols

These protocols all wrap some base message schema and provide
all the necessary hooks for pushing in a stream of bytes and
getting out packets in the order they were found.  The protocol
handlers will also provide notifications of error conditions
(for instance, unexpected bytes or a bad checksum).

"""
from functools import partial

import six
from suitcase.fields import Magic


class StreamProtocolHandler(object):
    """Protocol handler that deals fluidly with a stream of bytes

    The protocol handler is agnostic to the data source or methodology
    being used to collect the data  (blocking reads on a socket to async
    IO on a serial port).

    Here's an example of what one usage might look like (very simple
    approach for parsing a simple TCP protocol::


        from suitcase.protocol import StreamProtocolHandler
        from suitcase.fields import LengthField, UBInt16, VariableRawPayload
        from suitcase.structure import Structure
        import socket

        class SimpleFramedMessage(Structure):
            length = LengthField(UBInt16())
            payload = VariableRawPayload(length)

        def packet_received(packet):
            print(packet)

        def run_forever(host, port):
            protocol_handler = StreamProtocolHandler(SimpleFramedMessage,
                                                     packet_received)
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect(host, port)
            sock.setblocking(1)
            while True:
                bytes = sock.recv(1024)
                if len(bytes) == 0:
                    print("Socket closed... exiting")
                    return
                else:
                    protocol_handler.feed(bytes)

    :param message_schema: The top-level message schema that defines the
        packets for the protocol to be used.
    :param packet_callback: A callback to be executed with the form
        ``callback(packet)`` when a fully-formed packet is detected.

    """

    def __init__(self, message_schema, packet_callback):
        # configuration parameters
        self.message_schema = message_schema
        self.packet_callback = packet_callback

        # internal state
        self._available_bytes = b""
        self._packet_generator = self._create_packet_generator()

    def _create_packet_generator(self):
        while True:
            curmsg = self.message_schema()
            for i, (_name, field) in enumerate(curmsg):
                bytes_required = field.bytes_required

                if i == 0 and isinstance(field, Magic):
                    magic_seq = field.getval()
                    while True:
                        if len(self._available_bytes) < bytes_required:
                            yield None
                            continue

                        idx = self._available_bytes.find(magic_seq)
                        if idx == -1:  # no match in buffer
                            # Since we know the entire magic_seq is not here, there can be at most
                            # bytes_required - 1 bytes of the magic_seq available.  Thus we keep
                            # that many bytes around in case it is the start of the magic field.
                            self._available_bytes = self._available_bytes[-(bytes_required - 1):]
                            yield None
                        else:
                            self._available_bytes = self._available_bytes[idx:]
                            break  # continue processing

                # For a specific field, read until we have enough bytes
                # and then give the field a try.
                while True:
                    bytes_available = len(self._available_bytes)
                    if bytes_required <= bytes_available:
                        field_bytes = self._available_bytes[:bytes_required]
                        new_bytes = self._available_bytes[bytes_required:]
                        self._available_bytes = new_bytes
                        field.unpack(field_bytes)
                        break
                    else:
                        yield None
            yield curmsg

    def feed(self, new_bytes):
        """Feed a new set of bytes into the protocol handler

        These bytes will be immediately fed into the parsing state machine and
        if new packets are found, the ``packet_callback`` will be executed
        with the fully-formed message.

        :param new_bytes: The new bytes to be fed into the stream protocol
            handler.

        """
        self._available_bytes += new_bytes
        callbacks = []
        try:
            while True:
                packet = six.next(self._packet_generator)
                if packet is None:
                    break
                else:
                    callbacks.append(partial(self.packet_callback, packet))
        except Exception:
            # When we receive an exception, we assume that the _available_bytes
            # has already been updated and we just choked on a field.  That
            # is, unless the number of _available_bytes has not changed.  In
            # that case, we reset the buffered entirely

            # TODO: black hole may not be the best.  What should the logging
            # behavior be?
            self.reset()

        # callbacks are partials that are bound to packet already.  We do
        # this in order to separate out parsing activity (and error handling)
        # from the execution of callbacks.  Callbacks should not in any way
        # rely on the parsers position in the byte stream.
        for callback in callbacks:
            callback()

    def reset(self):
        """Reset the internal state machine to a fresh state

        If the protocol in use does not properly handle cases of possible
        de-synchronization it might be necessary to issue a reset if bytes
        are being received but no packets are coming out of the state
        machine.  A reset is issue internally whenever an unexpected exception
        is encountered while processing bytes from the stream.

        """
        self._packet_generator = self._create_packet_generator()
        self._available_bytes = b""



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2019 Digi International Inc. All Rights Reserved.

import unittest

import six
from suitcase.crc import crc16_ccitt, crc32
from suitcase.exceptions import SuitcaseProgrammingError, SuitcasePackStructException, SuitcasePackException, \
    SuitcaseParseError, SuitcaseException
from suitcase.fields import DependentField, LengthField, VariableRawPayload, \
    Magic, BitField, BitBool, BitNum, DispatchTarget, CRCField, Payload, \
    UBInt8, UBInt16, UBInt24, UBInt32, UBInt40, UBInt48, UBInt56, UBInt64, \
    SBInt8, SBInt16, SBInt24, SBInt32, SBInt40, SBInt48, SBInt56, SBInt64, \
    ULInt8, ULInt16, ULInt24, ULInt32, ULInt40, ULInt48, ULInt56, ULInt64, \
    SLInt8, SLInt16, SLInt24, SLInt32, SLInt40, SLInt48, SLInt56, SLInt64, \
    ConditionalField, UBInt8Sequence, SBInt8Sequence, FieldProperty, \
    DispatchField, FieldArray, TypeField, SubstructureField
from suitcase.structure import Structure
import struct


def raise_value_error(*args, **kwargs):
    raise ValueError("Artifically Raised ValueError")


class SuperChild(Structure):
    options = DependentField('options')
    ubseq = DependentField('ubseq')
    length = LengthField(DependentField('submessage_length'))

    remaining = VariableRawPayload(length)


# See https://github.com/digidotcom/python-suitcase/issues/18
class BGAPIFrame(Structure):
    head = LengthField(
            BitField(16,
                     is_event=BitBool(),
                     technology_type=BitNum(4),
                     length=BitNum(11)),
            get_length=lambda f: f.length,
            set_length=lambda f, l: setattr(f, 'length', l))
    class_id = UBInt8()
    command_id = UBInt8()
    payload = Payload(head)


class TestCustomSetGetLengthField(unittest.TestCase):

    def test_unpack(self):
        frame = BGAPIFrame.from_data(b'\x00\x00\x00\x01')
        self.assertEqual(frame.head.is_event, False)
        self.assertEqual(frame.head.technology_type, 0)
        self.assertEqual(frame.head.length, 0)
        self.assertEqual(frame.class_id, 0)
        self.assertEqual(frame.command_id, 1)
        self.assertEqual(frame.payload, b"")

    def test_pack(self):
        frame = BGAPIFrame()
        frame.head.is_event = True
        frame.head.technology_type = 0b1101
        frame.class_id = 0x10
        frame.command_id = 0xFE
        frame.payload = b"Hello, world!"
        self.assertEqual(frame.pack(), b'\xe8\r\x10\xfeHello, world!')
        frame_readback = BGAPIFrame.from_data(frame.pack())
        self.assertEqual(frame_readback.head.length, len(frame.payload))


# Message containing every field
class SuperMessage(Structure):
    magic = Magic(b'\xAA\xAA')

    # bitfield
    options = BitField(8,
                       b1=BitBool(),
                       b2=BitBool(),
                       rest=BitNum(6))

    # unsigned big endian
    ubint8 = UBInt8()
    ubint16 = UBInt16()
    ubint24 = UBInt24()
    ubint32 = UBInt32()
    ubint40 = UBInt40()
    ubint48 = UBInt48()
    ubint56 = UBInt56()
    ubint64 = UBInt64()

    # signed big endian
    sbint8 = SBInt8()
    sbint16 = SBInt16()
    sbint24 = SBInt24()
    sbint32 = SBInt32()
    sbint40 = SBInt40()
    sbint48 = SBInt48()
    sbint56 = SBInt56()
    sbint64 = SBInt64()

    # unsigned little endian
    ulint8 = ULInt8()
    ulint16 = ULInt16()
    ulint24 = ULInt24()
    ulint32 = ULInt32()
    ulint40 = ULInt40()
    ulint48 = ULInt48()
    ulint56 = ULInt56()
    ulint64 = ULInt64()

    # signed little endian
    slint8 = SLInt8()
    slint16 = SLInt16()
    slint24 = SLInt24()
    slint32 = SLInt32()
    slint40 = SLInt40()
    slint48 = SLInt48()
    slint56 = SLInt56()
    slint64 = SLInt64()

    # optional
    optional_one = ConditionalField(UBInt8(), lambda m: m.options.b1)
    optional_two = ConditionalField(UBInt8(), lambda m: m.options.b2)

    # sequences with variable lengths
    ubseql = LengthField(UBInt8())
    ubseq = UBInt8Sequence(ubseql)

    sbseql = LengthField(UBInt8())
    sbseq = SBInt8Sequence(sbseql)

    # sequences with fixed lengths
    ubseqf = UBInt8Sequence(5)
    sbseqf = SBInt8Sequence(5)

    # don't change anything... for test coverage
    ulint16_value = FieldProperty(ulint16)
    ulint16_byte_string = FieldProperty(ulint16,
                                        onget=lambda v: str(v),
                                        onset=lambda v: struct.unpack(">H", v)[0])

    message_type = DispatchField(UBInt8())
    submessage_length = LengthField(UBInt16())
    submessage = DispatchTarget(submessage_length, message_type, {
        0xEF: SuperChild
    })

    # checksum starts after beginning magic, ends before
    # the checksum
    crc = CRCField(UBInt16(), crc16_ccitt, 2, -3)
    eof = Magic(b'~')


class TestMagic(unittest.TestCase):
    def test_magic(self):
        magic_field = Magic('\xAA').create_instance(None)
        self.assertRaises(SuitcaseProgrammingError, magic_field.setval)


class TestSuperField(unittest.TestCase):
    def _create_supermessage(self):
        s = SuperMessage()

        s.options.b1 = False
        s.options.b2 = True
        s.options.remaining = 0x1A

        # packed fields
        s.ubint8 = 0xAA
        s.ubint16 = 0xAABB
        s.ubint24 = 0xAABBCC
        s.ubint32 = 0xAABBCCDD
        s.ubint40 = 0xAABBCCDDEE
        s.ubint48 = 0xAABBCCDDEEFF
        s.ubint56 = 0xAABBCCDDEEFF00
        s.ubint64 = 0xAABBCCDDEEFF0011

        s.sbint8 = -100
        s.sbint16 = -1000
        s.sbint24 = -100000
        s.sbint32 = -100000000
        s.sbint40 = -10000000000
        s.sbint48 = -10000000000000
        s.sbint56 = -1000000000000000
        s.sbint64 = -100000000000000000

        s.ulint8 = 0xAA
        s.ulint16 = 0xAABB
        s.ulint16_byte_string = b'\xAA\xBB'
        s.ulint16_value = 0xBBAA

        s.ulint24 = 0xAABBCC
        s.ulint32 = 0xAABBCCDD
        s.ulint40 = 0xAABBCCDDEE
        s.ulint48 = 0xAABBCCDDEEFF
        s.ulint56 = 0xAABBCCDDEEFF00
        s.ulint64 = 0xAABBCCDDEEFF0011

        s.slint8 = -100
        s.slint16 = -1000
        s.slint24 = -100000
        s.slint32 = -100000000
        s.slint40 = -10000000000
        s.slint48 = -10000000000000
        s.slint56 = -1000000000000000
        s.slint64 = -100000000000000000

        s.optional_one = 1
        s.optional_two = 2

        s.ubseq = tuple(range(10))
        s.sbseq = tuple([x - 5 for x in range(10)])
        s.ubseqf = tuple(range(5))
        s.sbseqf = tuple([x - 2 for x in range(5)])

        sub = SuperChild()
        sub.remaining = b"Hello, this is SuperChild!"
        s.submessage = sub

        return s

    def test_super_message(self):
        sm = self._create_supermessage()
        packed = sm.pack()
        sm2 = SuperMessage.from_data(packed)

        for key, field in sm2:
            sm2_value = field.getval()
            if key == 'options':
                self.assertEqual(sm2_value.b1, sm.options.b1)
                self.assertEqual(sm2_value.b2, sm.options.b2)
                self.assertEqual(sm2_value.rest, sm.options.rest)

            elif key == 'submessage':
                self.assertEqual(sm2_value.ubseq, sm.ubseq)
                self.assertEqual(sm2_value.remaining, sm.submessage.remaining)
            elif key == 'optional_one':
                self.assertEqual(sm2_value, None)
            elif key == 'crc':
                pass  # validity check baked into protocol
            else:
                sm1_value = getattr(sm, key)
                self.assertEqual(sm2_value, sm1_value,
                                 "%s: %s != %s, types(%s, %s)" % (
                                     key, sm2_value, sm1_value, type(sm2_value), type(sm1_value)))

    def test_struct_packing_exceptions(self):
        # Verify that exceptions available all work
        sm = self._create_supermessage()
        sm.ubint8 = 0xFFFF  # too big
        self.assertRaises(SuitcasePackStructException, sm.pack)

    def test_other_pack_exception(self):
        sm = self._create_supermessage()
        sm._key_to_field['ubint8'].pack = raise_value_error
        self.assertRaises(SuitcasePackException, sm.pack)

    def test_repr_works(self):
        # just make sure nothing crashes
        sm = self._create_supermessage()
        repr(sm)


class TestFieldProperty(unittest.TestCase):
    def test_basic_setget(self):
        # define the message
        class MyMessage(Structure):
            _version = UBInt8Sequence(2)
            version = FieldProperty(_version,
                                    onget=lambda v: "%d.%02d" % (v[0], v[1]),
                                    onset=lambda v: tuple(int(six.b(x)) for x in v.split(".", 1)))

        msg = MyMessage.from_data(b'\x10\x03')
        self.assertEqual(msg._version, (16, 3))
        self.assertEqual(msg.version, "16.03")

        msg.version = "22.7"
        self.assertEqual(msg._version, (22, 7))
        self.assertEqual(msg.version, "22.07")


class BasicMessage(Structure):
    b1 = UBInt8()
    b2 = UBInt8()


class TestInstancePrototyping(unittest.TestCase):
    def test_independence(self):
        msg1 = BasicMessage()
        msg1.b1 = 10
        msg1.b2 = 20

        msg2 = BasicMessage()
        msg2.b1 = 20
        msg2.b2 = 30

        self.assertNotEqual(msg2.b1, msg1.b1)
        self.assertNotEqual(msg2.b2, msg1.b2)


class TestLengthField(unittest.TestCase):
    class MyMuiltipliedLengthMessage(Structure):
        length = LengthField(UBInt8(), multiplier=8)
        payload = VariableRawPayload(length)

    class MyLengthyMessage(Structure):
        length = LengthField(UBInt16())
        payload = VariableRawPayload(length)

    def test_basic_length_pack(self):
        msg = self.MyLengthyMessage()
        payload = b"Hello, world!"
        msg.payload = payload
        self.assertEqual(msg.pack(), b"\x00\x0DHello, world!")

    def test_basic_length_unpack(self):
        msg = self.MyLengthyMessage()
        msg.unpack(b'\x00\x0DHello, world!')
        self.assertEqual(msg.length, 13)
        self.assertEqual(msg.payload, b"Hello, world!")

    def test_multiplied_length_pack(self):
        msg = self.MyMuiltipliedLengthMessage()
        payload = b''.join(six.b(chr(x)) for x in range(8 * 4))
        msg.payload = payload
        self.assertEqual(msg.pack(), b'\x04' + payload)

    def test_bad_modulus_multiplier(self):
        cls = self.MyMuiltipliedLengthMessage
        msg = self.MyMuiltipliedLengthMessage()
        payload = b'\x01'  # 1-byte is not modulo 8
        msg.payload = payload
        self.assertRaises(SuitcaseProgrammingError, msg.pack)

    def test_multiplied_length_unpack(self):
        msg = self.MyMuiltipliedLengthMessage()
        msg.unpack(b'\x04' + b''.join([six.b(chr(x)) for x in range(8 * 4)]))
        self.assertEqual(msg.length, 4)
        self.assertEqual(msg.payload,
                         b''.join([six.b(chr(x)) for x in range(8 * 4)]))

    def test_unpack_insufficient_bytes(self):
        msg = self.MyLengthyMessage()
        self.assertRaises(SuitcaseParseError, msg.unpack, b'\x00\x0DHello')


class TestByteSequence(unittest.TestCase):
    def test_fixed_sequence(self):
        class MySeqMessage(Structure):
            type = UBInt8()
            byte_values = UBInt8Sequence(16)

        msg = MySeqMessage()
        msg.type = 0
        msg.byte_values = (0, 1, 2, 3, 4, 5, 6, 7, 8,
                           9, 10, 11, 12, 13, 14, 15)
        self.assertEqual(msg.pack(),
                         b'\x00\x00\x01\x02\x03\x04\x05\x06\x07\x08'
                         b'\t\n\x0b\x0c\r\x0e\x0f')

    def test_variable_sequence(self):
        class MyVarSeqMessage(Structure):
            type = UBInt8()
            length = LengthField(UBInt8())
            seq = UBInt8Sequence(length)

        msg = MyVarSeqMessage()
        msg.type = 0
        msg.seq = (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)

        self.assertEqual(msg.pack(),
                         b'\x00\n\x01\x02\x03\x04\x05\x06\x07\x08\t\n')
        msg2 = MyVarSeqMessage()
        msg2.unpack(msg.pack())
        self.assertEqual(msg2.length, 10)
        self.assertEqual(msg2.seq, (1, 2, 3, 4, 5, 6, 7, 8, 9, 10))

    def test_variable_sequence_nolength(self):
        class MyVarSeqMessage(Structure):
            s = VariableRawPayload(None)
            b = UBInt8Sequence(None)

        m = MyVarSeqMessage()
        m.s = b"Hello, world - "
        m.b = [0, 1, 2, 3, 4, 5]
        self.assertEqual(m.pack(),
                         b'Hello, world - \x00\x01\x02\x03\x04\x05')


class MyTargetMessage(Structure):
    # inherited from the parent message
    _length = LengthField(DependentField('length'))
    payload = VariableRawPayload(_length)


class MyOtherTargetMessage(Structure):
    #: Inherited from the parent message
    _length = LengthField(DependentField('length'))
    sequence = UBInt8Sequence(_length)


class MySimpleFixedPayload(Structure):
    number = UBInt32()


class MyPayloadMessage(Structure):
    payload = Payload()


class MyDefaultTargetMessage(Structure):
    _length = LengthField(DependentField('length'))
    sequence = UBInt8Sequence(_length)


class MyBasicDispatchMessage(Structure):
    type = DispatchField(UBInt8())
    length = LengthField(UBInt16())
    body = DispatchTarget(length, type, {
        0x00: MyTargetMessage,
        0x01: MyOtherTargetMessage,
        None: MyDefaultTargetMessage
    })


class MyDynamicLengthDispatchMessage(Structure):
    type = DispatchField(UBInt8())
    body = DispatchTarget(None, type, {
        0x1F: MySimpleFixedPayload,
        0xCC: MyPayloadMessage,
    })
    eof = Magic(b'EOF')


class BasicGreedy(Structure):
    a = UBInt8()
    b = UBInt8()
    payload = Payload()


class BoxedGreedy(Structure):
    sof = Magic(b'\xAA')
    a = LengthField(UBInt8())
    b = UBInt8()
    payload = Payload()
    c = UBInt16()
    d = VariableRawPayload(a)
    eof = Magic(b'\xbb')


class CRCGreedyTail(Structure):
    payload = Payload()
    magic = Magic(b'~~')
    crc = CRCField(UBInt32(), crc32, 0, -1)  # all (checksum zeroed)


class TestGreedyFields(unittest.TestCase):
    def test_unpack_basic_greedy(self):
        # Test case with trailing greedy payload
        m = BasicGreedy()
        m.unpack(b"\x00\x01Hello, you greedy, greedy World!")
        self.assertEqual(m.a, 0x00)
        self.assertEqual(m.b, 0x01)
        self.assertEqual(m.payload, b"Hello, you greedy, greedy World!")

    def test_unpack_boxed_greedy(self):
        # Test case where fields exist on either side of payload
        m = BoxedGreedy()
        m.unpack(b"\xaa\x05\x00This is the payload\x00\x15ABCDE\xbb")
        self.assertEqual(m.a, 5)
        self.assertEqual(m.b, 0)
        self.assertEqual(m.payload, b"This is the payload")
        self.assertEqual(m.c, 0x15)
        self.assertEqual(m.d, b"ABCDE")

    def test_unpack_boxed_greedy_error_before_greedy(self):
        # should start with \xAA but does not
        m = BoxedGreedy()
        self.assertRaises(SuitcaseParseError, m.unpack,
                          b"\x11\x05\x00This is the payload\x00\x15ABCDE\xbb")
        m._key_to_field['sof'].unpack = raise_value_error  # simulate fault
        self.assertRaises(SuitcaseParseError, m.unpack,
                          b"\xaa\x05\x00This is the payload\x00\x15ABCDE\xbb")

    def test_unpack_boxed_greedy_error_after_greedy(self):
        # should end in \xBB but does not
        m = BoxedGreedy()
        self.assertRaises(SuitcaseParseError, m.unpack,
                          b"\xaa\x05\x00This is the payload\x00\x15ABCDE\xAF")

        m._key_to_field['eof'].unpack = raise_value_error  # simlulate fault
        self.assertRaises(SuitcaseParseError, m.unpack,
                          b"\xaa\x05\x00This is the payload\x00\x15ABCDE\xbb")

    def test_pack_basic_greedy(self):
        m = BasicGreedy()
        m.a = 10
        m.b = 20
        m.payload = b"This is a packed greedy payload"
        self.assertEqual(m.pack(), b'\n\x14This is a packed greedy payload')

    def test_pack_boxed_greedy(self):
        m = BoxedGreedy()
        m.b = 20
        m.c = 300
        m.d = b"ABCD"
        m.payload = b"My length isn't declared and I am in the middle"
        self.assertEqual(m.pack(),
                         b"\xaa\x04\x14My length isn't declared and I am in the "
                         b"middle\x01,ABCD\xbb")

    def test_greedy_tail_pack_basic(self):
        m = CRCGreedyTail()
        m.payload = b"A GREEDY PAYLOAD FROM THE START"
        self.assertEqual(m.pack(), b'A GREEDY PAYLOAD FROM THE START~~\xfc\xce`=')

    def test_greedy_tail_unpack_basic(self):
        m = CRCGreedyTail.from_data(b'A GREEDY PAYLOAD FROM THE START~~\xfc\xce`=')
        self.assertEqual(m.payload, b"A GREEDY PAYLOAD FROM THE START")

    def test_greedy_tail_not_enough_bytes(self):
        self.assertRaises(SuitcaseParseError, CRCGreedyTail.from_data, b'~\xfc\xce`=')


class TestMessageDispatching(unittest.TestCase):
    def test_dispatch_packing(self):
        msg = MyBasicDispatchMessage()
        target_msg = MyTargetMessage()
        target_msg.payload = b"Hello, world!"
        msg.body = target_msg
        self.assertEqual(msg.pack(), b"\x00\x00\rHello, world!")

        msg2 = MyBasicDispatchMessage()
        target_msg = MyOtherTargetMessage()
        target_msg.sequence = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
        msg2.body = target_msg
        self.assertEqual(msg2.pack(),
                         b'\x01\x00\x0b\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n')

    def test_dispatch_unpacking(self):
        msg = MyBasicDispatchMessage()
        msg.unpack(b"\x00\x00\rHello, world!")
        self.assertEqual(msg.type, 0x00)
        body = msg.body
        self.assertEqual(body.payload, b"Hello, world!")

    def test_foreign_type(self):
        msg = MyBasicDispatchMessage()
        self.assertRaises(SuitcaseProgrammingError,
                          setattr, msg, 'body', SuperMessage())

    def test_default_dispatch(self):
        msg = MyBasicDispatchMessage()
        msg.unpack(b"\x10\x00\rHello, world!")
        self.assert_(isinstance(msg.body, MyDefaultTargetMessage))


class TestMessageDispatchingVariableLength(unittest.TestCase):
    def test_fixed_field_dispatch_packing(self):
        msg = MyDynamicLengthDispatchMessage()
        target_msg = MySimpleFixedPayload()
        target_msg.number = 0xFFEEFFEE
        msg.body = target_msg
        self.assertEqual(msg.pack(), b"\x1f\xff\xee\xff\xeeEOF")

    def test_variable_field_dispatch_packing(self):
        msg = MyDynamicLengthDispatchMessage()
        target_msg = MyPayloadMessage()
        target_msg.payload = b"~~~I'M SO DYNAMIC~~~"
        msg.body = target_msg
        self.assertEqual(msg.pack(), b"\xcc~~~I'M SO DYNAMIC~~~EOF")

    def test_dispatch_unpacking_fixed(self):
        msg = MyDynamicLengthDispatchMessage.from_data(b"\x1f\xff\xee\xff\xeeEOF")
        self.assertEqual(msg.body.number, 0xFFEEFFEE)

    def test_dispatch_unpacking_variable(self):
        msg = MyDynamicLengthDispatchMessage.from_data(b"\xcc~~~I'M SO DYNAMIC~~~EOF")
        self.assertEqual(msg.body.payload, b"~~~I'M SO DYNAMIC~~~")

    def test_dispatch_unpacking_fixed_bad_length(self):
        # too many bytes before EOF
        self.assertRaises(SuitcaseParseError, MyDynamicLengthDispatchMessage.from_data,
                          b"\x1f\xff\xee\xff\xeeEXTRA_EOF")


class TestBitFields(unittest.TestCase):
    def test_packing(self):
        field_proto = BitField(16,
                               nib1=BitNum(4),
                               nib2=BitNum(4),
                               nib3=BitNum(4),
                               nib4=BitNum(4),
                               )
        field = field_proto.create_instance(None)

        field.nib1 = 1
        field.nib2 = 2
        field.nib3 = 3
        field.nib4 = 4

        field2 = field_proto.create_instance(None)
        sio = six.BytesIO()
        field.pack(sio)
        field2.unpack(sio.getvalue())
        self.assertEqual(field2.nib1, 1)
        self.assertEqual(field2.nib2, 2)
        self.assertEqual(field2.nib3, 3)
        self.assertEqual(field2.nib4, 4)

    def test_bad_operations(self):
        field_proto = BitField(7,
                               num=BitNum(7))
        self.assertRaises(SuitcaseProgrammingError,
                          field_proto.create_instance, None)

    def test_explicit_field_override(self):
        field_proto = BitField(16, ULInt16(),
                               b1=BitBool(),
                               b2=BitBool(),
                               remaining=BitNum(14))
        inst = field_proto.create_instance(None)
        inst.b1 = True
        inst.b2 = False
        inst.remaining = 0x1EF
        sio = six.BytesIO()
        inst.pack(sio)

        # should be packed in little endian form
        self.assertEqual(sio.getvalue(), b"\xef\x81")

        inst2 = field_proto.create_instance(None)
        inst2.unpack(b"\xef\x81")
        self.assertEqual(inst.b1, inst2.b1)
        self.assertEqual(inst.b2, inst2.b2)
        self.assertEqual(inst.remaining, inst2.remaining)


# message where f2 is only defined if f2 is 255
class Conditional(Structure):
    f1 = UBInt8()
    f2 = ConditionalField(UBInt8(), lambda m: m.f1 == 255)


# message where f2 and f3 are only defined if f2 is 255,
# and f2 specifies a length for f3
class ConditionalLength(Structure):
    f1 = UBInt8()
    f2 = ConditionalField(LengthField(UBInt8()), lambda m: m.f1 == 255)
    f3 = ConditionalField(Payload(length_provider=f2), lambda m: m.f1 == 255)


# message body that is empty
class EmptyStructure(Structure):
    pass


# message where f3 is only defined if f2 is 255,
# and f3 is a dispatch target based on f1
class ConditionalDispatch(Structure):
    f1 = DispatchField(UBInt8())
    f2 = UBInt8()  # separate from f1
    length = LengthField(UBInt8())
    f3 = ConditionalField(
            DispatchTarget(length, f1, {0: EmptyStructure, 1: BasicMessage}),
            condition=lambda m: m.f2 == 255)


class ConditionalSubstructures(Structure):
    f1 = UBInt8()
    f2 = ConditionalField(SubstructureField(BasicMessage),
                          condition=lambda m: m.f1 < 10)
    f3 = ConditionalField(SubstructureField(BasicMessage),
                          condition=lambda m: m.f1 < 20)


class TestConditionalField(unittest.TestCase):
    def test_conditional_pack(self):
        m1 = Conditional()
        m1.f1 = 0x91
        self.assertEqual(m1.pack(), b'\x91')

        m2 = Conditional()
        m2.f1 = 0xFF
        m2.f2 = 0x09
        self.assertEqual(m2.pack(), b'\xff\x09')

    def test_conditional_rx(self):
        m1 = Conditional()
        m1.unpack(b'\x1f')
        self.assertEqual(m1.f1, 0x1f)
        self.assertEqual(m1.f2, None)

        m2 = Conditional()
        m2.unpack(b'\xff\x1f')
        self.assertEqual(m2.f1, 0xff)
        self.assertEqual(m2.f2, 0x1f)

    def test_conditional_length_pack(self):
        m1 = ConditionalLength()
        m1.f1 = 0x91
        self.assertEqual(m1.pack(), b'\x91')

        m2 = ConditionalLength()
        m2.f1 = 0xFF
        m2.f3 = b'\x01\x02\x03\x04\x05'
        self.assertEqual(m2.pack(), b'\xff\x05\x01\x02\x03\x04\x05')

    def test_conditional_length_unpack(self):
        m1 = ConditionalLength()
        m1.unpack(b'\x91')
        self.assertEqual(m1.f1, 0x91)
        self.assertEqual(m1.f2, None)
        self.assertEqual(m1.f3, None)

        m2 = ConditionalLength()
        m2.unpack(b'\xff\x05\x01\x02\x03\x04\x05')
        self.assertEqual(m2.f1, 0xff)
        self.assertEqual(m2.f2, 0x05)
        self.assertEqual(m2.f3, b'\x01\x02\x03\x04\x05')

    def test_conditional_dispatch(self):
        # Dispatch to an empty structure
        m1 = ConditionalDispatch()
        m1.unpack(b'\x00\xff\x00')
        self.assertEqual(m1.f1, 0)
        self.assertEqual(m1.f2, 255)
        self.assertEqual(m1.length, 0)
        self.assertIsInstance(m1.f3, EmptyStructure)

        # Dispatch to a non-empty structure
        m2 = ConditionalDispatch()
        m2.unpack(b'\x01\xff\x02\x11\x22')
        self.assertEqual(m2.f1, 1)
        self.assertEqual(m2.f2, 255)
        self.assertEqual(m2.length, 2)
        self.assertIsInstance(m2.f3, BasicMessage)
        self.assertEqual(m2.f3.b1, 0x11)
        self.assertEqual(m2.f3.b2, 0x22)

    def test_conditional_substructure_pack(self):
        m = ConditionalSubstructures()
        m.f1 = 9

        f2 = BasicMessage()
        f2.b1 = 0x55
        f2.b2 = 0x66
        m.f2 = f2

        f3 = BasicMessage()
        f3.b1 = 0x77
        f3.b2 = 0x88
        m.f3 = f3

        self.assertEqual(m.pack(), b"\x09\x55\x66\x77\x88")

        # Remove f2
        m.f1 = 10
        m.f2 = None
        self.assertEqual(m.pack(), b"\x0a\x77\x88")

        # Remove f3
        m.f1 = 20
        m.f3 = None
        self.assertEqual(m.pack(), b"\x14")

    def test_conditional_substructure_unpack(self):
        # Neither substructure is present
        m1 = ConditionalSubstructures.from_data(b"\x20")
        self.assertEqual(m1.f1, 0x20)
        self.assertEqual(m1.f2, None)
        self.assertEqual(m1.f3, None)

        # Include f3 but not f2
        m2 = ConditionalSubstructures()
        m2.unpack(b"\x10" +
                  b"\x11\x22")
        self.assertEqual(m2.f1, 0x10)
        self.assertEqual(m2.f2, None)
        self.assertIsInstance(m2.f3, BasicMessage)
        self.assertEqual(m2.f3.b1, 0x11)
        self.assertEqual(m2.f3.b2, 0x22)

        # Include both substructures
        m3 = ConditionalSubstructures()
        m3.unpack(b"\x00" +
                  b"\x11\x22" +
                  b"\x33\x44")
        self.assertEqual(m3.f1, 0x00)
        self.assertIsInstance(m3.f2, BasicMessage)
        self.assertEqual(m3.f2.b1, 0x11)
        self.assertEqual(m3.f2.b2, 0x22)
        self.assertIsInstance(m3.f3, BasicMessage)
        self.assertEqual(m3.f3.b1, 0x33)
        self.assertEqual(m3.f3.b2, 0x44)


class TestStructure(unittest.TestCase):
    def test_unpack_fewer_bytes_than_required(self):
        self.assertRaises(SuitcaseParseError, MySimpleFixedPayload.from_data, b'123')

    def test_unpack_more_bytes_than_required(self):
        self.assertRaises(SuitcaseParseError, MySimpleFixedPayload.from_data, b'12345')


# Test FieldArray and ConditionalField interaction
class ConditionalArrayElement(Structure):
    options = BitField(8,
                       cond8_present=BitBool(),
                       cond16_present=BitBool(),
                       unused=BitNum(6))
    cond8 = ConditionalField(UBInt8(), lambda m: m.options.cond8_present)
    cond16 = ConditionalField(UBInt16(), lambda m: m.options.cond16_present)


class ConditionalArrayGreedy(Structure):
    array = FieldArray(ConditionalArrayElement)


class ConditionalArrayGreedyAfter(Structure):
    length = LengthField(UBInt8())
    array = FieldArray(ConditionalArrayElement, length)
    greedy = Payload()


class TestConditionalArray(unittest.TestCase):
    def test_unpack_greedy(self):
        m = ConditionalArrayGreedy()
        m.unpack(b"\x00" +
                 b"\x80\x12" +
                 b"\x40\x34\x56" +
                 b"\xc0\x12\x34\x56")

        self.assertEqual(len(m.array), 4)

        self.assertEqual(m.array[0].options.cond8_present, 0)
        self.assertEqual(m.array[0].options.cond16_present, 0)
        self.assertEqual(m.array[0].cond8, None)
        self.assertEqual(m.array[0].cond16, None)

        self.assertEqual(m.array[1].options.cond8_present, 1)
        self.assertEqual(m.array[1].options.cond16_present, 0)
        self.assertEqual(m.array[1].cond8, 0x12)
        self.assertEqual(m.array[1].cond16, None)

        self.assertEqual(m.array[2].options.cond8_present, 0)
        self.assertEqual(m.array[2].options.cond16_present, 1)
        self.assertEqual(m.array[2].cond8, None)
        self.assertEqual(m.array[2].cond16, 0x3456)

        self.assertEqual(m.array[3].options.cond8_present, 1)
        self.assertEqual(m.array[3].options.cond16_present, 1)
        self.assertEqual(m.array[3].cond8, 0x12)
        self.assertEqual(m.array[3].cond16, 0x3456)

    def test_pack_greedy(self):
        m = ConditionalArrayGreedy()

        c1 = ConditionalArrayElement()
        c1.options.cond8_present = False
        c1.options.cond16_present = False
        c1.cond8 = None
        c1.cond16 = None
        m.array.append(c1)

        c2 = ConditionalArrayElement()
        c2.options.cond8_present = True
        c2.options.cond16_present = False
        c2.cond8 = 0x12
        c2.cond16 = None
        m.array.append(c2)

        c3 = ConditionalArrayElement()
        c3.options.cond8_present = False
        c3.options.cond16_present = True
        c3.cond8 = None
        c3.cond16 = 0x3456
        m.array.append(c3)

        c4 = ConditionalArrayElement()
        c4.options.cond8_present = True
        c4.options.cond16_present = True
        c4.cond8 = 0x12
        c4.cond16 = 0x3456
        m.array.append(c4)

        self.assertEqual(m.pack(),
                         b"\x00" +
                         b"\x80\x12" +
                         b"\x40\x34\x56" +
                         b"\xc0\x12\x34\x56")

    def test_unpack_greedy_after(self):
        m = ConditionalArrayGreedyAfter()
        m.unpack(b"\x05" +
                 b"\x00" +
                 b"\xc0\x12\x34\x56" +
                 b"Hello")

        self.assertEqual(len(m.array), 2)

        self.assertEqual(m.array[0].options.cond8_present, 0)
        self.assertEqual(m.array[0].options.cond16_present, 0)
        self.assertEqual(m.array[0].cond8, None)
        self.assertEqual(m.array[0].cond16, None)

        self.assertEqual(m.array[1].options.cond8_present, 1)
        self.assertEqual(m.array[1].options.cond16_present, 1)
        self.assertEqual(m.array[1].cond8, 0x12)
        self.assertEqual(m.array[1].cond16, 0x3456)

        self.assertEqual(m.greedy, b"Hello")

    def test_pack_greedy_after(self):
        m = ConditionalArrayGreedyAfter()

        c1 = ConditionalArrayElement()
        c1.options.cond8_present = False
        c1.options.cond16_present = False
        c1.cond8 = None
        c1.cond16 = None
        m.array.append(c1)

        c2 = ConditionalArrayElement()
        c2.options.cond8_present = True
        c2.options.cond16_present = True
        c2.cond8 = 0x12
        c2.cond16 = 0x3456
        m.array.append(c2)

        m.greedy = b"Hello"

        self.assertEqual(m.pack(),
                         b"\x05" +
                         b"\x00" +
                         b"\xc0\x12\x34\x56" +
                         b"Hello")


# Test TypeField and FieldArray interaction
class Structure8(Structure):
    value = UBInt8()


class Structure16(Structure):
    value = UBInt16()


class Structure32(Structure):
    value = UBInt32()


dispatch_mapping = {0x00: Structure8,
                    0x01: Structure16,
                    0x03: Structure32}

size_mapping = {0x00: 1,
                0x01: 2,
                None: 4}  # This handles the Structure32 case


class TypedArrayElement(Structure):
    type = TypeField(UBInt8(), size_mapping)
    value = DispatchTarget(type, type, dispatch_mapping)


class TypedArray(Structure):
    array = FieldArray(TypedArrayElement)


class TypeFieldNoLengthProvider(Structure):
    type = TypeField(UBInt8(), size_mapping)


class TypeFieldWrongSize(Structure):
    type = TypeField(UBInt8(), size_mapping)
    greedy = Payload(type)


class TestTypeField(unittest.TestCase):
    def test_unpack_typed_array(self):
        m = TypedArray.from_data(b"\x00\x12" +
                                 b"\x01\x12\x34" +
                                 b"\x03\x12\x34\x56\x78")

        self.assertEqual(m.array[0].type, 0x00)
        self.assertEqual(m.array[0].value.value, 0x12)

        self.assertEqual(m.array[1].type, 0x01)
        self.assertEqual(m.array[1].value.value, 0x1234)

        self.assertEqual(m.array[2].type, 0x03)
        self.assertEqual(m.array[2].value.value, 0x12345678)

    def test_pack_typed_array(self):
        m = TypedArray()

        e1 = TypedArrayElement()
        e1.value = Structure8.from_data(b"\x12")
        m.array.append(e1)

        e2 = TypedArrayElement()
        e2.value = Structure16.from_data(b"\x12\x34")
        m.array.append(e2)

        e3 = TypedArrayElement()
        e3.value = Structure32.from_data(b"\x12\x34\x56\x78")
        m.array.append(e3)

        self.assertEquals(m.pack(), b"\x00\x12" +
                          b"\x01\x12\x34" +
                          b"\x03\x12\x34\x56\x78")

    def test_repr(self):
        # Make sure nothing crashes
        m = TypedArray.from_data(b"\x03\x12\x34\x56\x78")
        repr(m)

    def test_no_length_provider(self):
        m = TypeFieldNoLengthProvider()
        m.type = 0x01
        self.assertRaises(SuitcaseException, m.pack)

    def test_wrong_size(self):
        m = TypeFieldWrongSize()
        m.type = 0x01  # Indicates a length of 2
        m.greedy = b"Hello"
        self.assertRaises(SuitcaseProgrammingError, m.pack)


# Test empty FieldArray
class BasicMessageArray(Structure):
    count = LengthField(UBInt8(), multiplier=2)
    array = FieldArray(BasicMessage, count)


# Test empty FieldArray with content after the array
class BasicMessageArrayAfter(Structure):
    count = LengthField(UBInt8(), multiplier=2)
    array = FieldArray(BasicMessage, count)
    after = UBInt8()


# Test a FieldArray that takes a number of elements instead of a size in bytes.
class BasicMessageArrayNumElements(Structure):
    count = LengthField(UBInt8())
    array = FieldArray(BasicMessage, num_elements_provider=count)


class TestFieldArray(unittest.TestCase):
    def test_pack_valid(self):
        m = BasicMessageArray()

        first = BasicMessage()
        first.b1 = 0x11
        first.b2 = 0x22
        second = BasicMessage()
        second.b1 = 0x33
        second.b2 = 0x44

        m.array = [first, second]

        self.assertEqual(m.pack(), b"\x02\x11\x22\x33\x44")

    def test_unpack_valid(self):
        m = BasicMessageArray.from_data(b"\x01AB")
        self.assertEqual(m.count, 1)
        self.assertEqual(len(m.array), 1)
        self.assertIsInstance(m.array[0], BasicMessage)
        self.assertEqual(m.array[0].b1, ord('A'))
        self.assertEqual(m.array[0].b2, ord('B'))

    def test_pack_empty(self):
        m = BasicMessageArray()

        m.array = []

        self.assertEqual(m.pack(), b"\x00")

    def test_unpack_empty(self):
        m = BasicMessageArray.from_data(b"\x00")
        self.assertEqual(m.count, 0)
        self.assertEqual(len(m.array), 0)

    def test_pack_after_valid(self):
        m = BasicMessageArrayAfter()

        m.array = []
        m.after = 1

        self.assertEqual(m.pack(), b"\x00\x01")

        # Populate the array.
        m2 = BasicMessageArrayAfter()
        first = BasicMessage()
        first.b1 = 0x22
        first.b2 = 0x33
        m2.array = [first]
        m2.after = 0

        self.assertEquals(m2.pack(), b"\x01\x22\x33\x00")

    def test_unpack_after_valid(self):
        m = BasicMessageArrayAfter.from_data(b"\x00\x11")

        self.assertEqual(m.count, 0)
        self.assertEqual(len(m.array), 0)
        self.assertEqual(m.after, 0x11)

        m2 = BasicMessageArrayAfter.from_data(b"\x01\x11\x22\x00")

        self.assertEqual(m2.count, 1)
        self.assertEqual(len(m2.array), 1)
        self.assertIsInstance(m2.array[0], BasicMessage)
        self.assertEqual(m2.array[0].b1, 0x11)
        self.assertEqual(m2.array[0].b2, 0x22)
        self.assertEqual(m2.after, 0)

    def test_pack_num_elements_valid(self):
        m = BasicMessageArrayNumElements()
        m.array = []
        self.assertEqual(m.pack(), b"\x00")

        messages = [BasicMessage(), BasicMessage(), BasicMessage()]
        for i, message in enumerate(messages):
            message.b1 = 0x10 + i
            message.b2 = 0x20 + i
            m.array.append(message)

        self.assertEqual(m.pack(), b"\x03\x10\x20\x11\x21\x12\x22")

    def test_unpack_num_elements_valid(self):
        m = BasicMessageArrayNumElements.from_data(b"\x02\x22\x33\x44\x55")

        self.assertEqual(m.count, 2)
        self.assertEqual(len(m.array), 2)
        self.assertIsInstance(m.array[0], BasicMessage)
        self.assertEqual(m.array[0].b1, 0x22)
        self.assertEqual(m.array[0].b2, 0x33)
        self.assertEqual(m.array[1].b1, 0x44)
        self.assertEqual(m.array[1].b2, 0x55)


# Test SubstructureField
class PascalString16(Structure):
    length = LengthField(UBInt16())
    value = Payload(length)


class NameStructure(Structure):
    first = SubstructureField(PascalString16)
    last = SubstructureField(PascalString16)


class NameStructureGreedyAfter(Structure):
    first = SubstructureField(PascalString16)
    last = SubstructureField(PascalString16)
    greedy = Payload()


class TestSubstructureField(unittest.TestCase):
    def test_pack_valid(self):
        m = NameStructure()
        m.first.value = b"John"
        m.last.value = b"Doe"
        self.assertEqual(m.pack(), b"\x00\x04John\x00\x03Doe")

    def test_unpack_valid(self):
        m = NameStructure().from_data(b"\x00\x04John\x00\x03Doe")
        self.assertIsInstance(m.first, PascalString16)
        self.assertEqual(m.first.value, b"John")
        self.assertIsInstance(m.last, PascalString16)
        self.assertEqual(m.last.value, b"Doe")

    def test_pack_greedy_after(self):
        m = NameStructureGreedyAfter()
        m.first.value = b"John"
        m.last.value = b"Doe"
        m.greedy = b"Hello World!"
        self.assertEqual(m.pack(), b"\x00\x04John\x00\x03DoeHello World!")

    def test_unpack_greedy_after(self):
        m = NameStructureGreedyAfter().from_data(b"\x00\x04John\x00\x03DoeHello World!")
        self.assertIsInstance(m.first, PascalString16)
        self.assertEqual(m.first.value, b"John")
        self.assertIsInstance(m.last, PascalString16)
        self.assertEqual(m.last.value, b"Doe")
        self.assertEqual(m.greedy, b"Hello World!")


class TestDiscoverableFields(unittest.TestCase):
    def test_dir(self):
        m = PascalString16()
        attrs = dir(m)
        self.assertTrue("length" in attrs)
        self.assertTrue("value" in attrs)


class TestKeywordArgumentInitialization(unittest.TestCase):
    def test_basic_assignment(self):
        m = PascalString16(value=b"Hello World!")
        self.assertEqual(m.pack(), b"\x00\x0cHello World!")

    def test_nested_assignment(self):
        m = NameStructure(first=PascalString16(value=b"John"),
                          last=PascalString16(value=b"Doe"))
        self.assertEqual(m.pack(), b"\x00\x04John\x00\x03Doe")

    def test_partial_assignment(self):
        m = NameStructure(first=PascalString16(value=b"John"))
        m.last = PascalString16(value=b"Doe")
        self.assertEqual(m.pack(), b"\x00\x04John\x00\x03Doe")


class StructureWithFieldAccessorAtTopLevel(Structure):
    bits = BitField(8, top=BitNum(3), bottom=BitNum(5))
    top = bits.top


class TestFieldAccessorAtTopLevel(unittest.TestCase):
    def test_access(self):
        s = StructureWithFieldAccessorAtTopLevel.from_data(b'\xA0')
        # Sanity check.
        self.assertEqual(s.bits.top, 0b101)
        self.assertEqual(s.bits.bottom, 0b00000)
        self.assertEqual(s.top, 0b101)

    def test_assignment(self):
        s = StructureWithFieldAccessorAtTopLevel.from_data(b'\x00')
        s.top = 0b111
        self.assertEqual(s.pack(), b'\xE0')


class Zero(Structure):
    body = Payload()


class One(Structure):
    first = UBInt8()
    second = UBInt16()


class FieldAccessorDispatch(Structure):
    bits = BitField(8, upper=BitNum(5), lower=BitNum(3))
    body = DispatchTarget(None, bits.upper, {
        0: Zero,
        1: One,
    })


class TestFieldAccessorDispatch(unittest.TestCase):
    def test_on_unpack(self):
        z = FieldAccessorDispatch.from_data(b'\x07Hello')
        self.assert_(isinstance(z.body, Zero))
        self.assertEqual(z.body.body, b'Hello')

        o = FieldAccessorDispatch.from_data(b'\x08\x11\x22\x33')
        self.assert_(isinstance(o.body, One))
        self.assertEqual(o.body.first, 0x11)
        self.assertEqual(o.body.second, 0x2233)

    def test_assignment(self):
        zero = FieldAccessorDispatch.from_data(b'\x00Hello there')
        self.assertEqual(zero.pack(), b'\x00Hello there')

        zero.body = One(first=1, second=2)
        self.assertEqual(zero.pack(), b'\x08\x01\x00\x02')


class FieldAccessorMultiple(Structure):
    # FieldAccessors can be referenced multiple times,
    # and more than one attribute of a given field may be referenced.
    # TODO: Figure out how to support LengthField(bits.upper)?
    bits = BitField(8, upper=BitNum(4), lower=BitNum(4))
    upper = bits.upper
    lower = bits.lower
    body = DispatchTarget(None, bits.lower, {
        0b1010: Structure16,  # 0x0A
        0b0101: One,  # 0x05
    })


class TestFieldAccessorMultiple(unittest.TestCase):
    def test_unpack(self):
        s = FieldAccessorMultiple.from_data(b'\x5A\xEE\x00')
        self.assert_(isinstance(s.body, Structure16))
        self.assertEqual(s.body.value, 0xEE00)

        s2 = FieldAccessorMultiple.from_data(
                b'\x35\xFF\x00\x01')
        self.assert_(isinstance(s2.body, One))
        self.assertEqual(s2.body.first, 0xFF)
        self.assertEqual(s2.body.second, 0x0001)

    def test_assignment(self):
        s = FieldAccessorMultiple(body=One(first=2, second=3))
        self.assertEqual(s.lower, 0b0101)
        self.assertEqual(s.body.first, 2)
        self.assertEqual(s.body.second, 3)
        self.assertEqual(s.pack(), b'\x05\x02\x00\x03')

        s.body = Structure16(value=0xFFFE)
        s.upper = 0b1100  # 0xC
        self.assertEqual(s.bits.lower, 0b1010)
        self.assertEqual(s.lower, s.bits.lower)
        self.assertEqual(s.upper, s.bits.upper)
        self.assertEqual(s.pack(), b'\xCA\xFF\xFE')


class _InnerDispatchType0(Structure):
    some_field = UBInt8()
    cond_sub = SubstructureField(ConditionalSubstructures)


class _InnerDispatchType1(Structure):
    one = UBInt8()
    two = UBInt8()


class _OuterDispatch(Structure):
    which = DispatchField(UBInt8())
    # NOTE: This setup will fail if greedy=False is not specified,
    # indicating that _OuterDispatch didn't consume the remainder of the data.
    # Because we don't _expect_ it to consume the rest of the data in the message,
    # we need to mark this as non-greedy.
    # TODO: We may be able to infer this from field ordering, or apply some heuristic
    # over the entire structure to correctly find the greedy field (payload at the end,
    # in this case).
    dispatch = DispatchTarget(None, which, {
        0x00: _InnerDispatchType0,
        0x01: _InnerDispatchType1,
    }, greedy=False)


class SubstructureDispatchFollowedByPayload(Structure):
    some_value = UBInt8()
    sub = SubstructureField(_OuterDispatch)
    another_value = UBInt8()
    payload = Payload()


class TestSubstructureDispatchFollowedByPayload(unittest.TestCase):
    def test_unpack(self):
        s = SubstructureDispatchFollowedByPayload.from_data(
            (
                # some_value
                b'\xF7'
                # sub: _OuterDispatch
                # - which
                b'\x00'
                # - _InnerDispatchType0
                # -- some_field
                b'\x7F'
                # -- cond_sub: ConditionalSubstructures
                # --- f1=19.
                b'\x13'
                # --- f3, 2 bytes.
                b'\x77\x88'
                # another_value
                b'\xE5'
                # payload
                b'This is my payload'
            )
        )

        self.assertEqual(s.some_value, 0xF7)
        self.assertEqual(s.sub.which, 0)
        self.assertEqual(s.sub.dispatch.some_field, 0x7F)
        self.assertEqual(s.sub.dispatch.cond_sub.f1, 19)
        self.assertEqual(s.sub.dispatch.cond_sub.f3.b1, 0x77)
        self.assertEqual(s.sub.dispatch.cond_sub.f3.b2, 0x88)
        self.assertEqual(s.another_value, 0xE5)
        self.assertEqual(s.payload, b'This is my payload')

    def test_pack(self):
        sub_dispatch = _InnerDispatchType0(
            some_field=0x01,
            cond_sub=ConditionalSubstructures(
                f1=11,
                f3=BasicMessage(
                    b1=0x01,
                    b2=0x02,
                )
            )
        )
        structure = SubstructureDispatchFollowedByPayload()
        structure.some_value = 0xFE
        structure.another_value = 0xE5
        structure.payload = b"Hello World!"
        structure.sub.dispatch = sub_dispatch

        self.assertEqual(structure.pack(),
                         b"\xFE\x00\x01\x0B\x01\x02\xE5Hello World!")


class FieldArrayWithDependentField(Structure):
    length = UBInt8()
    # Each entry in the array has a payload of length `length`
    entries = FieldArray(MyTargetMessage)


class SubstructureWithDependentField(Structure):
    length = UBInt16()
    sub = SubstructureField(MyTargetMessage)


class TestFieldArrayWithDependentField(unittest.TestCase):
    def test_unpack(self):
        data = (
            b"\x05"  # each message is 5 bytes
            b"Hello"
            b"World"
            b"Digi!"
        )
        structure = FieldArrayWithDependentField.from_data(data)
        self.assertEqual(len(structure.entries), 3)
        self.assertEqual(5, structure.length)
        self.assertEqual(structure.entries[0].payload, b"Hello")
        self.assertEqual(structure.entries[1].payload, b"World")
        self.assertEqual(structure.entries[2].payload, b"Digi!")

    @unittest.expectedFailure
    def test_dependent_field_automatically_updates(self):
        structure = FieldArrayWithDependentField()
        structure.entries.append(MyTargetMessage(payload=b"ABC"))
        self.assertEqual(3, structure.length)


class TestSubstructureWithDependentField(unittest.TestCase):
    def test_unpack(self):
        data = (
            b"\x00\x08"
            b"Hello..."
        )
        structure = SubstructureWithDependentField.from_data(data)
        self.assertEqual(structure.length, 8)
        self.assertEqual(structure.sub.payload, b"Hello...")

    @unittest.expectedFailure
    def test_auto_assignment_of_dependent_field(self):
        data = b"\x00\x01A"
        structure = SubstructureWithDependentField.from_data(data)
        structure.sub.payload = b"Hello, World!"
        # We would probably expect this to work! But it does not. :(
        self.assertEqual(structure.length, 13)


class MultipleGreedyFields(Structure):
    # TODO: Ideally we could do this particular example?
    payload1 = Payload()
    magic = Magic(b'\x00')
    payload2 = Payload()


class TestMultipleGreedyFields(unittest.TestCase):
    def test_unpack(self):
        with self.assertRaises(SuitcaseParseError) as caught:
            MultipleGreedyFields.from_data(
                b"Hello\0World"
            )
        self.assertIn(
            "found another greedy field", str(caught.exception),
            "Suitcase is not expected to support multiple greedy fields")

    def test_pack(self):
        # Even though we can't unpack such a structure, we can construct
        # and pack it.
        structure = MultipleGreedyFields(
            payload1=b"Hello,",
            payload2=b"World!",
        )
        self.assertEqual(structure.pack(), b"Hello,\x00World!")


class GreedyPayloadFollowedByNongreedyDispatch(Structure):
    payload = Payload()
    _nul = Magic(b'\x00')
    kind = DispatchField(UBInt8())
    rest = DispatchTarget(None, kind, {
        0x00: BasicMessage,
        0x01: MySimpleFixedPayload,
    }, greedy=False)


class TestGreedyPayloadFollowedByNongreedyDispatch(unittest.TestCase):
    def test_cannot_unpack(self):
        data = (
            b"Hello, World!"
            b"\x00"
            b"\x01"
            b"\xDE\xAD\xBE\xEF"
        )
        # Suitcase does not know how to handle a greedy field that is followed
        # by a nongreedy dispatch target. (If the length of the dispatch target
        # is indeterminate, the unpack process will feed the entire message in,
        # which obviously is not what we want.) The solution to this would be
        # to be able to direct Suitcase to parse the fields left-to-right, since
        # in this case the greedy payload is "boxed" by a null terminator.
        # But if that terminator was not present, this structure would be
        # impossible to parse anyway.
        self.assertRaises(SuitcaseParseError,
                          GreedyPayloadFollowedByNongreedyDispatch.from_data,
                          data)

    def test_pack(self):
        structure = GreedyPayloadFollowedByNongreedyDispatch()
        structure.payload = b"Hello!"
        structure.rest = MySimpleFixedPayload(number=0xDEADBEEF)
        self.assertEqual(structure.pack(), b"Hello!\x00\x01\xDE\xAD\xBE\xEF")


if __name__ == "__main__":
    unittest.main()
